import contextlib
import inspect
import json
import logging
import shutil
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from datetime import date, datetime, time
from decimal import Decimal
from os import listdir, scandir, unlink
from os.path import isfile, join
from pathlib import Path, PosixPath, PureWindowsPath
from typing import Any, Callable, Dict, Generator, List, Optional, Tuple, Type, Union

import ruamel.yaml
import yaml
from pydantic import SecretStr, ValidationError
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefinedType
from yaml import SafeDumper

from hummingbot import get_strategy_list, root_path
from hummingbot.client.config.client_config_map import ClientConfigMap
from hummingbot.client.config.config_data_types import BaseClientModel, ClientConfigEnum, ClientFieldData
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.config.fee_overrides_config_map import fee_overrides_config_map, init_fee_overrides_config
from hummingbot.client.config.gateway_ssl_config_map import SSLConfigMap
from hummingbot.client.settings import (
    CLIENT_CONFIG_PATH,
    CONF_DIR_PATH,
    CONF_POSTFIX,
    CONF_PREFIX,
    CONNECTORS_CONF_DIR_PATH,
    GATEWAY_SSL_CONF_FILE,
    STRATEGIES_CONF_DIR_PATH,
    TEMPLATE_PATH,
    TRADE_FEES_CONFIG_PATH,
    AllConnectorSettings,
)


class ConfigValidationError(Exception):
    pass


@dataclass()
class ConfigTraversalItem:
    depth: int
    config_path: str
    attr: str
    value: Any
    printable_value: str
    client_field_data: Optional[ClientFieldData]
    field_info: FieldInfo
    type_: Type


class ClientConfigAdapter:
    def __init__(self, hb_config: BaseClientModel):
        self._hb_config = hb_config

    def __getattr__(self, item):
        value = getattr(self._hb_config, item)
        if isinstance(value, BaseClientModel):
            value = ClientConfigAdapter(value)
        return value

    def __setattr__(self, key, value):
        if key == "_hb_config":
            super().__setattr__(key, value)
        else:
            try:
                self._hb_config.__setattr__(key, value)
            except ValidationError as e:
                raise ConfigValidationError(retrieve_validation_error_msg(e))

    def __repr__(self):
        return f"{self.__class__.__name__}.{self._hb_config.__repr__()}"

    def __eq__(self, other):
        if isinstance(other, ClientConfigAdapter):
            eq = self._hb_config.__eq__(other._hb_config)
        else:
            eq = super().__eq__(other)
        return eq

    @property
    def hb_config(self) -> BaseClientModel:
        return self._hb_config

    @property
    def title(self) -> str:
        return self._hb_config.model_config["title"]

    def is_required(self, attr: str) -> bool:
        return self._hb_config.is_required(attr)

    def keys(self) -> Generator[str, None, None]:
        return self._hb_config.__class__.model_fields.keys()

    def config_paths(self) -> Generator[str, None, None]:
        return (traversal_item.config_path for traversal_item in self.traverse())

    def traverse(self, secure: bool = True) -> Generator[ConfigTraversalItem, None, None]:
        """The intended use for this function is to simplify config map traversals in the client code.

        If the field is missing, its value will be set to `None` and its printable value will be set to
        'MISSING_AND_REQUIRED'.
        """
        depth = 0
        for attr, field_info in self._hb_config.__class__.model_fields.items():
            type_ = field_info.annotation
            if hasattr(self, attr):
                value = getattr(self, attr)
                printable_value = self._get_printable_value(attr, value, secure)
                client_field_data = self.get_client_data(attr)
            else:
                value = None
                printable_value = "&cMISSING_AND_REQUIRED"
                client_field_data = self.get_client_data(attr)
            yield ConfigTraversalItem(
                depth=depth,
                config_path=attr,
                attr=attr,
                value=value,
                printable_value=printable_value,
                client_field_data=client_field_data,
                field_info=field_info,
                type_=type_,
            )
            if isinstance(value, ClientConfigAdapter):
                for traversal_item in value.traverse():
                    traversal_item.depth += 1
                    config_path = f"{attr}.{traversal_item.config_path}"
                    traversal_item.config_path = config_path
                    yield traversal_item

    async def get_client_prompt(self, attr_name: str) -> Optional[str]:
        prompt = None
        client_data = self.get_client_data(attr_name)
        if client_data is not None:
            prompt_fn = client_data.prompt
            if inspect.iscoroutinefunction(prompt_fn):
                prompt = await prompt_fn(self._hb_config)
            elif inspect.isfunction(prompt_fn):
                prompt = prompt_fn(self._hb_config)
            elif isinstance(prompt_fn, str):
                prompt = prompt_fn
        return prompt

    def is_secure(self, attr_name: str) -> bool:
        client_data = self.get_client_data(attr_name)
        secure = client_data is not None and client_data.is_secure
        return secure

    def get_client_data(self, attr_name: str) -> Optional[ClientFieldData]:
        json_schema_extra = self._hb_config.__class__.model_fields[attr_name].json_schema_extra or {}
        client_data = ClientFieldData(
            prompt=json_schema_extra.get("prompt"),
            prompt_on_new=json_schema_extra.get("prompt_on_new", False),
            is_secure=json_schema_extra.get("is_secure", False),
            is_connect_key=json_schema_extra.get("is_connect_key", False),
            is_updatable=json_schema_extra.get("is_updatable", False),
        )
        return client_data

    def get_description(self, attr_name: str) -> str:
        return self._hb_config.__class__.model_fields[attr_name].description

    def get_default(self, attr_name: str) -> Any:
        default = self._hb_config.__class__.model_fields[attr_name].default
        if isinstance(default, type(Ellipsis)) or isinstance(default, PydanticUndefinedType):
            default = None
        return default

    def get_default_str_repr(self, attr_name: str) -> str:
        """Used to generate default strings for config prompts."""
        default = self.get_default(attr_name=attr_name)
        if default is None:
            default_str = ""
        elif isinstance(default, (List, Tuple)):
            default_str = ",".join(default)
        elif isinstance(default, BaseClientModel):
            default_str = default.model_config["title"]
        else:
            default_str = str(default)
        return default_str

    def get_type(self, attr_name: str) -> Type:
        return self._hb_config.__class__.model_fields[attr_name].annotation

    def generate_yml_output_str_with_comments(self) -> str:
        fragments_with_comments = [self._generate_title()]
        self._add_model_fragments(fragments_with_comments)
        yml_str = "".join(fragments_with_comments)
        return yml_str

    def setattr_no_validation(self, attr: str, value: Any):
        with self._disable_validation():
            setattr(self, attr, value)

    def full_copy(self):
        return self.__class__(hb_config=self._hb_config.model_copy(deep=True))

    def decrypt_all_secure_data(self):
        from hummingbot.client.config.security import Security  # avoids circular import

        secure_config_items = (
            traversal_item
            for traversal_item in self.traverse()
            if traversal_item.client_field_data is not None and traversal_item.client_field_data.is_secure
        )
        for traversal_item in secure_config_items:
            value = traversal_item.value
            if isinstance(value, SecretStr):
                value = value.get_secret_value()
            if value == "" or Security.secrets_manager is None:
                decrypted_value = value
            else:
                decrypted_value = Security.secrets_manager.decrypt_secret_value(attr=traversal_item.attr, value=value)
            *intermediate_items, final_config_element = traversal_item.config_path.split(".")
            config_model = self
            if len(intermediate_items) > 0:
                for attr in intermediate_items:
                    config_model = config_model.__getattr__(attr)
            setattr(config_model, final_config_element, decrypted_value)

    @contextlib.contextmanager
    def _disable_validation(self):
        self._hb_config.model_config["validate_assignment"] = False
        yield
        self._hb_config.model_config["validate_assignment"] = True

    def _get_printable_value(self, attr: str, value: Any, secure: bool) -> str:
        if isinstance(value, ClientConfigAdapter):
            if self._is_union(self.get_type(attr)):  # it is a union of modes
                printable_value = value.hb_config.model_config["title"]
            else:  # it is a collection of settings stored in a submodule
                printable_value = ""
        elif isinstance(value, SecretStr) and not secure:
            printable_value = value.get_secret_value()
        else:
            printable_value = str(value)
        return printable_value

    @staticmethod
    def _is_union(t: Type) -> bool:
        is_union = hasattr(t, "__origin__") and t.__origin__ == Union
        return is_union

    def _dict_in_conf_order(self) -> Dict[str, Any]:
        conf_dict = {}
        for attr in self._hb_config.__class__.model_fields.keys():
            value = getattr(self, attr)
            if isinstance(value, ClientConfigAdapter):
                value = value._dict_in_conf_order()
            conf_dict[attr] = value
        self._encrypt_secrets(conf_dict)
        return conf_dict

    def _encrypt_secrets(self, conf_dict: Dict[str, Any]):
        from hummingbot.client.config.security import Security  # avoids circular import
        for attr, value in conf_dict.items():
            if isinstance(value, SecretStr):
                clear_text_value = value.get_secret_value() if isinstance(value, SecretStr) else value
                if not Security.secrets_manager:
                    logging.getLogger().warning(f"Ignore the following error if your config file {attr} contains secret(s)")
                conf_dict[attr] = Security.secrets_manager.encrypt_secret_value(attr, clear_text_value)

    def _decrypt_secrets(self, conf_dict: Dict[str, Any]):
        from hummingbot.client.config.security import Security  # avoids circular import
        for attr, value in conf_dict.items():
            attr_type = self._hb_config.model_fields[attr].annotation
            if attr_type == SecretStr:
                decrypted_value = Security.secrets_manager.decrypt_secret_value(attr, value.get_secret_value())
                conf_dict[attr] = SecretStr(decrypted_value)

    def _decrypt_all_internal_secrets(self):
        from hummingbot.client.config.security import Security  # avoids circular import

        for traversal_item in self.traverse():
            if traversal_item.type_ == SecretStr:
                encrypted_value = traversal_item.value
                if isinstance(encrypted_value, SecretStr):
                    encrypted_value = encrypted_value.get_secret_value()
                decrypted_value = Security.secrets_manager.decrypt_secret_value(traversal_item.attr, encrypted_value)
                parent_attributes = traversal_item.config_path.split(".")[:-1]
                config = self
                for parent_attribute in parent_attributes:
                    config = getattr(config, parent_attribute)
                setattr(config, traversal_item.attr, decrypted_value)

    def _generate_title(self) -> str:
        title = f"{self._hb_config.model_config['title']}"
        title = self._adorn_title(title)
        return title

    @staticmethod
    def _adorn_title(title: str) -> str:
        if title:
            title = f"###   {title} config   ###"
            title_len = len(title)
            title = f"{'#' * title_len}\n{title}\n{'#' * title_len}"
        return title

    def _add_model_fragments(
        self,
        fragments_with_comments: List[str],
    ):

        fragments_with_comments.append("\n")
        first_level_conf_items_generator = (item for item in self.traverse() if item.depth == 0)

        for traversal_item in first_level_conf_items_generator:
            fragments_with_comments.append("\n")

            attr_comment = traversal_item.field_info.description
            if attr_comment is not None:
                comment_prefix = f"{' ' * 2 * traversal_item.depth}# "
                attr_comment = "\n".join(f"{comment_prefix}{c}" for c in attr_comment.split("\n"))
                fragments_with_comments.append(attr_comment)
                fragments_with_comments.append("\n")

            attribute = traversal_item.attr
            value = getattr(self, attribute)
            if isinstance(value, ClientConfigAdapter):
                value = value._dict_in_conf_order()
            if isinstance(traversal_item.value, PureWindowsPath):
                conf_as_dictionary = {attribute: traversal_item.printable_value}
            else:
                conf_as_dictionary = {attribute: value}
            self._encrypt_secrets(conf_as_dictionary)

            yaml_config = yaml.safe_dump(conf_as_dictionary, sort_keys=False)
            fragments_with_comments.append(yaml_config)


class ReadOnlyClientConfigAdapter(ClientConfigAdapter):
    def __setattr__(self, key, value):
        if key == "_hb_config":
            super().__setattr__(key, value)
        else:
            raise AttributeError("Cannot set an attribute on a read-only client adapter")

    @classmethod
    def lock_config(cls, config_map: ClientConfigMap):
        return cls(config_map._hb_config)


# Use ruamel.yaml to preserve order and comments in .yml file
yaml_parser = ruamel.yaml.YAML()  # legacy


def decimal_representer(dumper: SafeDumper, data: Decimal):
    return dumper.represent_float(float(data))


def enum_representer(dumper: SafeDumper, data: ClientConfigEnum):
    return dumper.represent_str(str(data))


def date_representer(dumper: SafeDumper, data: date):
    return dumper.represent_date(data)


def time_representer(dumper: SafeDumper, data: time):
    return dumper.represent_str(data.strftime("%H:%M:%S"))


def datetime_representer(dumper: SafeDumper, data: datetime):
    return dumper.represent_datetime(data)


def path_representer(dumper: SafeDumper, data: Path):
    return dumper.represent_str(str(data))


def client_config_adapter_representer(dumper: SafeDumper, data: ClientConfigAdapter):
    return dumper.represent_dict(data._dict_in_conf_order())


def base_client_model_representer(dumper: SafeDumper, data: BaseClientModel):
    dictionary_representation = ClientConfigAdapter(data)._dict_in_conf_order()
    return dumper.represent_dict(dictionary_representation)


yaml.add_representer(
    data_type=Decimal, representer=decimal_representer, Dumper=SafeDumper
)
yaml.add_multi_representer(
    data_type=ClientConfigEnum, multi_representer=enum_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=date, representer=date_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=time, representer=time_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=datetime, representer=datetime_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=Path, representer=path_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=PosixPath, representer=path_representer, Dumper=SafeDumper
)
yaml.add_representer(
    data_type=ClientConfigAdapter, representer=client_config_adapter_representer, Dumper=SafeDumper
)
yaml.add_multi_representer(
    data_type=BaseClientModel, multi_representer=base_client_model_representer, Dumper=SafeDumper
)


def parse_cvar_value(cvar: ConfigVar, value: Any) -> Any:
    """
    Based on the target type specified in `ConfigVar.type_str`, parses a string value into the target type.
    :param cvar: ConfigVar object
    :param value: User input from running session or from saved `yml` files. Type is usually string.
    :return: value in the correct type
    """
    if value is None:
        return None
    elif cvar.type == 'str':
        return str(value)
    elif cvar.type == 'list':
        if isinstance(value, str):
            if len(value) == 0:
                return []
            filtered: filter = filter(lambda x: x not in ['[', ']', '"', "'"], list(value))
            value = "".join(filtered).split(",")  # create csv and generate list
            return [s.strip() for s in value]  # remove leading and trailing whitespaces
        else:
            return value
    elif cvar.type == 'json':
        if isinstance(value, str):
            value_json = value.replace("'", '"')  # replace single quotes with double quotes for valid JSON
            cvar_value = json.loads(value_json)
        else:
            cvar_value = value
        return cvar_json_migration(cvar, cvar_value)
    elif cvar.type == 'float':
        try:
            return float(value)
        except Exception:
            logging.getLogger().error(f"\"{value}\" is not valid float.", exc_info=True)
            return value
    elif cvar.type == 'decimal':
        try:
            return Decimal(str(value))
        except Exception:
            logging.getLogger().error(f"\"{value}\" is not valid decimal.", exc_info=True)
            return value
    elif cvar.type == 'int':
        try:
            return int(value)
        except Exception:
            logging.getLogger().error(f"\"{value}\" is not an integer.", exc_info=True)
            return value
    elif cvar.type == 'bool':
        if isinstance(value, str) and value.lower() in ["true", "yes", "y"]:
            return True
        elif isinstance(value, str) and value.lower() in ["false", "no", "n"]:
            return False
        else:
            return value
    else:
        raise TypeError


def cvar_json_migration(cvar: ConfigVar, cvar_value: Any) -> Any:
    """
    A special function to migrate json config variable when its json type changes, for paper_trade_account_balance
    and min_quote_order_amount (deprecated), they were List but change to Dict.
    """
    if cvar.key in ("paper_trade_account_balance", "min_quote_order_amount") and isinstance(cvar_value, List):
        results = {}
        for item in cvar_value:
            results[item[0]] = item[1]
        return results
    return cvar_value


def parse_cvar_default_value_prompt(cvar: ConfigVar) -> str:
    """
    :param cvar: ConfigVar object
    :return: text for default value prompt
    """
    if cvar.default is None:
        default = ""
    elif callable(cvar.default):
        default = cvar.default()
    elif cvar.type == 'bool' and isinstance(cvar.prompt, str) and "Yes/No" in cvar.prompt:
        default = "Yes" if cvar.default else "No"
    else:
        default = str(cvar.default)
    if isinstance(default, Decimal):
        default = "{0:.4f}".format(default)
    return default


async def copy_strategy_template(strategy: str) -> str:
    """
    Look up template `.yml` file for a particular strategy in `hummingbot/templates` and copy it to the `conf` folder.
    The file name is `conf_{STRATEGY}_strategy_{INDEX}.yml`
    :return: The newly created file name
    """
    old_path = get_strategy_template_path(strategy)
    i = 0
    new_fname = f"{CONF_PREFIX}{strategy}{CONF_POSTFIX}_{i}.yml"
    new_path = STRATEGIES_CONF_DIR_PATH / new_fname
    while isfile(new_path):
        new_fname = f"{CONF_PREFIX}{strategy}{CONF_POSTFIX}_{i}.yml"
        new_path = STRATEGIES_CONF_DIR_PATH / new_fname
        i += 1
    shutil.copy(old_path, new_path)
    return new_fname


def get_strategy_template_path(strategy: str) -> Path:
    """
    Given the strategy name, return its template config `yml` file name.
    """
    return TEMPLATE_PATH / f"{CONF_PREFIX}{strategy}{CONF_POSTFIX}_TEMPLATE.yml"


def _merge_dicts(*args: Dict[str, ConfigVar]) -> OrderedDict:
    """
    Helper function to merge a few dictionaries into an ordered dictionary.
    """
    result: OrderedDict[any] = OrderedDict()
    for d in args:
        result.update(d)
    return result


def get_connector_class(connector_name: str) -> Callable:
    conn_setting = AllConnectorSettings.get_connector_settings()[connector_name]
    mod = __import__(conn_setting.module_path(),
                     fromlist=[conn_setting.class_name()])
    return getattr(mod, conn_setting.class_name())


def get_strategy_config_map(
    strategy: str
) -> Optional[Union[ClientConfigAdapter, Dict[str, ConfigVar]]]:
    """
    Given the name of a strategy, find and load strategy-specific config map.
    """
    try:
        config_cls = get_strategy_pydantic_config_cls(strategy)
        if config_cls is None:  # legacy
            cm_key = f"{strategy}_config_map"
            strategy_module = __import__(f"hummingbot.strategy.{strategy}.{cm_key}",
                                         fromlist=[f"hummingbot.strategy.{strategy}"])
            config_map = getattr(strategy_module, cm_key)
        else:
            hb_config = config_cls.model_construct()
            config_map = ClientConfigAdapter(hb_config)
    except Exception:
        config_map = defaultdict()
    return config_map


def get_strategy_starter_file(strategy: str) -> Callable:
    """
    Given the name of a strategy, find and load the `start` function in
    `hummingbot/strategy/{STRATEGY_NAME}/start.py` file.
    """
    if strategy is None:
        return lambda: None
    try:
        strategy_module = __import__(f"hummingbot.strategy.{strategy}.start",
                                     fromlist=[f"hummingbot.strategy.{strategy}"])
        return getattr(strategy_module, "start")
    except Exception as e:
        logging.getLogger().error(e, exc_info=True)


def strategy_name_from_file(file_path: Path) -> str:
    data = read_yml_file(file_path)
    strategy = data.get("strategy")
    return strategy


def connector_name_from_file(file_path: Path) -> str:
    data = read_yml_file(file_path)
    connector = data["connector"]
    return connector


def validate_strategy_file(file_path: Path) -> Optional[str]:
    if not isfile(file_path):
        return f"{file_path} file does not exist."
    strategy = strategy_name_from_file(file_path)
    if strategy is None:
        return "Invalid configuration file or 'strategy' field is missing."
    if strategy not in get_strategy_list():
        return "Invalid strategy specified in the file."
    return None


def read_yml_file(yml_path: Path) -> Dict[str, Any]:
    with open(yml_path, "r", encoding="utf-8") as file:
        data = yaml.safe_load(file) or {}
    return dict(data)


def get_strategy_pydantic_config_cls(strategy_name: str):
    pydantic_cm_class = None
    try:
        pydantic_cm_pkg = f"{strategy_name}_config_map_pydantic"
        pydantic_cm_path = root_path() / "hummingbot" / "strategy" / strategy_name / f"{pydantic_cm_pkg}.py"
        if pydantic_cm_path.exists():
            pydantic_cm_class_name = f"{''.join([s.capitalize() for s in strategy_name.split('_')])}ConfigMap"
            pydantic_cm_mod = __import__(f"hummingbot.strategy.{strategy_name}.{pydantic_cm_pkg}",
                                         fromlist=[f"{pydantic_cm_class_name}"])
            pydantic_cm_class = getattr(pydantic_cm_mod, pydantic_cm_class_name)
    except ImportError:
        logging.getLogger().exception(f"Could not import Pydantic configs for {strategy_name}.")
    return pydantic_cm_class


async def load_strategy_config_map_from_file(yml_path: Path) -> Union[ClientConfigAdapter, Dict[str, ConfigVar]]:
    strategy_name = strategy_name_from_file(yml_path)
    config_cls = get_strategy_pydantic_config_cls(strategy_name)
    if config_cls is None:  # legacy
        config_map = get_strategy_config_map(strategy_name)
        template_path = get_strategy_template_path(strategy_name)
        await load_yml_into_cm_legacy(str(yml_path), str(template_path), config_map)
    else:
        config_data = read_yml_file(yml_path)
        hb_config = config_cls(**config_data)
        config_map = ClientConfigAdapter(hb_config)
    return config_map


def load_connector_config_map_from_file(yml_path: Path) -> ClientConfigAdapter:
    config_data = read_yml_file(yml_path)
    connector_name = connector_name_from_file(yml_path)
    hb_config = get_connector_hb_config(connector_name).model_validate(config_data)
    config_map = ClientConfigAdapter(hb_config)
    config_map.decrypt_all_secure_data()
    return config_map


def load_client_config_map_from_file() -> ClientConfigAdapter:
    yml_path = CLIENT_CONFIG_PATH
    if yml_path.exists():
        config_data = read_yml_file(yml_path)
    else:
        config_data = {}
    client_config = ClientConfigMap(**config_data)
    config_map = ClientConfigAdapter(client_config)
    save_to_yml(yml_path, config_map)
    return config_map


def load_ssl_config_map_from_file() -> ClientConfigAdapter:
    yml_path = GATEWAY_SSL_CONF_FILE
    if yml_path.exists():
        config_data = read_yml_file(yml_path)
    else:
        config_data = {}
    ssl_config = SSLConfigMap(**config_data)
    config_map = ClientConfigAdapter(ssl_config)

    if yml_path.exists():
        save_to_yml(yml_path, config_map)

    return config_map


def get_connector_hb_config(connector_name: str) -> BaseClientModel:
    hb_config = AllConnectorSettings.get_connector_config_keys(connector_name)
    return hb_config


def reset_connector_hb_config(connector_name: str):
    AllConnectorSettings.reset_connector_config_keys(connector_name)


def update_connector_hb_config(connector_config: ClientConfigAdapter):
    AllConnectorSettings.update_connector_config_keys(connector_config.hb_config)


def api_keys_from_connector_config_map(cm: ClientConfigAdapter) -> Dict[str, str]:
    api_keys = {}
    for c in cm.traverse():
        if c.value is not None and c.client_field_data is not None and c.client_field_data.is_connect_key:
            value = c.value.get_secret_value() if isinstance(c.value, SecretStr) else c.value
            api_keys[c.attr] = value
    return api_keys


def get_connector_config_yml_path(connector_name: str) -> Path:
    connector_path = Path(CONNECTORS_CONF_DIR_PATH) / f"{connector_name}.yml"
    return connector_path


def list_connector_configs() -> List[Path]:
    connector_configs = [
        Path(f.path) for f in scandir(str(CONNECTORS_CONF_DIR_PATH))
        if f.is_file() and not f.name.startswith("_") and not f.name.startswith(".")
    ]
    return connector_configs


async def load_yml_into_dict(yml_path: str) -> Dict[str, Any]:
    data = {}
    if isfile(yml_path):
        with open(yml_path, encoding="utf-8") as stream:
            data = yaml_parser.load(stream) or {}

    return dict(data.items())


async def save_yml_from_dict(yml_path: str, conf_dict: Dict[str, Any]):
    try:
        with open(yml_path, "w+", encoding="utf-8") as stream:
            data = yaml_parser.load(stream) or {}
            for key in conf_dict:
                data[key] = conf_dict.get(key)
            with open(yml_path, "w+", encoding="utf-8") as outfile:
                yaml_parser.dump(data, outfile)
    except Exception as e:
        logging.getLogger().error(f"Error writing configs: {str(e)}", exc_info=True)


async def load_yml_into_cm_legacy(yml_path: str, template_file_path: str, cm: Dict[str, ConfigVar]):
    try:
        data = {}
        conf_version = -1
        if isfile(yml_path):
            with open(yml_path, encoding="utf-8") as stream:
                data = yaml_parser.load(stream) or {}
                conf_version = data.get("template_version", 0)

        with open(template_file_path, "r", encoding="utf-8") as template_fd:
            template_data = yaml_parser.load(template_fd)
            template_version = template_data.get("template_version", 0)

        for key in template_data:
            if key in {"template_version"}:
                continue

            cvar = cm.get(key)
            if cvar is None:
                logging.getLogger().error(f"Cannot find corresponding config to key {key} in template.")
                continue

            if cvar.is_secure:
                raise DeprecationWarning("Secure values are no longer supported in legacy configs.")

            val_in_file = data.get(key, None)
            if (val_in_file is None or val_in_file == "") and cvar.default is not None:
                cvar.value = cvar.default
                continue

            # Todo: the proper process should be first validate the value then assign it
            cvar.value = parse_cvar_value(cvar, val_in_file)
            if cvar.value is not None:
                err_msg = await cvar.validate(str(cvar.value))
                if err_msg is not None:
                    # Instead of raising an exception, simply skip over this variable and wait till the user is prompted
                    logging.getLogger().error(
                        "Invalid value %s for config variable %s: %s" % (val_in_file, cvar.key, err_msg)
                    )
                    cvar.value = None

        if conf_version < template_version:
            # delete old config file
            if isfile(yml_path):
                unlink(yml_path)
            # copy the new file template
            shutil.copy(template_file_path, yml_path)
            # save the old variables into the new config file
            save_to_yml_legacy(yml_path, cm)
    except Exception as e:
        logging.getLogger().error("Error loading configs. Your config file may be corrupt. %s" % (e,),
                                  exc_info=True)


async def read_system_configs_from_yml():
    """
    Read global config and selected strategy yml files and save the values to corresponding config map
    If a yml file is outdated, it gets reformatted with the new template
    """
    await load_yml_into_cm_legacy(
        str(TRADE_FEES_CONFIG_PATH), str(TEMPLATE_PATH / "conf_fee_overrides_TEMPLATE.yml"), fee_overrides_config_map
    )
    # In case config maps get updated (due to default values)
    save_system_configs_to_yml()


def save_system_configs_to_yml():
    save_to_yml_legacy(str(TRADE_FEES_CONFIG_PATH), fee_overrides_config_map)


async def refresh_trade_fees_config(client_config_map: ClientConfigAdapter):
    """
    Refresh the trade fees config, after new connectors have been added (e.g. gateway connectors).
    """
    init_fee_overrides_config()
    save_to_yml(CLIENT_CONFIG_PATH, client_config_map)
    save_to_yml_legacy(str(TRADE_FEES_CONFIG_PATH), fee_overrides_config_map)


def save_to_yml_legacy(yml_path: str, cm: Dict[str, ConfigVar]):
    """
    Write current config saved a single config map into each a single yml file
    """
    try:
        with open(yml_path, encoding="utf-8") as stream:
            data = yaml_parser.load(stream) or {}
            for key in cm:
                cvar = cm.get(key)
                if isinstance(cvar.value, Decimal):
                    data[key] = float(cvar.value)
                else:
                    data[key] = cvar.value
            with open(yml_path, "w+", encoding="utf-8") as outfile:
                yaml_parser.dump(data, outfile)
    except Exception as e:
        logging.getLogger().error("Error writing configs: %s" % (str(e),), exc_info=True)


def save_to_yml(yml_path: Path, cm: ClientConfigAdapter):
    try:
        cm_yml_str = cm.generate_yml_output_str_with_comments()
        with open(yml_path, "w", encoding="utf-8") as outfile:
            outfile.write(cm_yml_str)
    except Exception as e:
        logging.getLogger().error("Error writing configs: %s" % (str(e),), exc_info=True)


def write_config_to_yml(
    strategy_config_map: Union[ClientConfigAdapter, Dict],
    strategy_file_name: str,
    client_config_map: ClientConfigAdapter,
):
    strategy_file_path = Path(STRATEGIES_CONF_DIR_PATH) / strategy_file_name
    if isinstance(strategy_config_map, ClientConfigAdapter):
        save_to_yml(strategy_file_path, strategy_config_map)
    else:
        save_to_yml_legacy(strategy_file_path, strategy_config_map)
    save_to_yml(CLIENT_CONFIG_PATH, client_config_map)


async def create_yml_files_legacy():
    """
    Copy `hummingbot_logs.yml` and `conf_global.yml` templates to the `conf` directory on start up
    """
    for fname in listdir(TEMPLATE_PATH):
        if "_TEMPLATE" in fname and CONF_POSTFIX not in fname:
            stripped_fname = fname.replace("_TEMPLATE", "")
            template_path = str(TEMPLATE_PATH / fname)
            conf_path = join(CONF_DIR_PATH, stripped_fname)
            if not isfile(conf_path):
                shutil.copy(template_path, conf_path)

            # Only overwrite log config. Updating `conf_global.yml` is handled by `read_configs_from_yml`
            if conf_path.endswith("hummingbot_logs.yml"):
                with open(template_path, "r", encoding="utf-8") as template_fd:
                    template_data = yaml_parser.load(template_fd)
                    template_version = template_data.get("template_version", 0)
                with open(conf_path, "r", encoding="utf-8") as conf_fd:
                    conf_version = 0
                    try:
                        conf_data = yaml_parser.load(conf_fd)
                        conf_version = conf_data.get("template_version", 0)
                    except Exception:
                        pass
                if conf_version < template_version:
                    shutil.copy(template_path, conf_path)


def default_strategy_file_path(strategy: str) -> str:
    """
    Find the next available file name.
    :return: a default file name - `conf_{short_strategy}_{INDEX}.yml` e.g. 'conf_pure_mm_1.yml'
    """
    i = 1
    new_fname = f"{CONF_PREFIX}{short_strategy_name(strategy)}_{i}.yml"
    new_path = STRATEGIES_CONF_DIR_PATH / new_fname
    while new_path.is_file():
        new_fname = f"{CONF_PREFIX}{short_strategy_name(strategy)}_{i}.yml"
        new_path = STRATEGIES_CONF_DIR_PATH / new_fname
        i += 1
    return new_fname


def short_strategy_name(strategy: str) -> str:
    if strategy == "pure_market_making":
        return "pure_mm"
    elif strategy == "cross_exchange_market_making":
        return "xemm"
    elif strategy == "arbitrage":
        return "arb"
    else:
        return strategy


def all_configs_complete(strategy_config: Union[ClientConfigAdapter, Dict], client_config_map: ClientConfigAdapter):
    return (
        config_map_complete_legacy(strategy_config)
        if isinstance(strategy_config, Dict)
        else True
    )


def config_map_complete_legacy(config_map):
    return not any(c.required and c.value is None for c in config_map.values())


def missing_required_configs_legacy(config_map):
    return [c for c in config_map.values() if c.required and c.value is None and not c.is_connect_key]


def format_config_file_name(file_name):
    if "." not in file_name:
        return file_name + ".yml"
    return file_name


def parse_config_default_to_text(config: ConfigVar) -> str:
    """
    :param config: ConfigVar object
    :return: text for default value prompt
    """
    if config.default is None:
        default = ""
    elif callable(config.default):
        default = config.default()
    elif config.type == 'bool' and isinstance(config.prompt, str) and "Yes/No" in config.prompt:
        default = "Yes" if config.default else "No"
    else:
        default = str(config.default)
    if isinstance(default, Decimal):
        default = "{0:.4f}".format(default)
    return default


def retrieve_validation_error_msg(e: ValidationError) -> str:
    return e.errors().pop()["msg"]



import asyncio
from collections import defaultdict
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from async_timeout import timeout

from hummingbot.connector.client_order_tracker import ClientOrderTracker
from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.injective_v2 import (
    injective_constants as CONSTANTS,
    injective_v2_web_utils as web_utils,
)
from hummingbot.connector.exchange.injective_v2.injective_events import InjectiveEvent
from hummingbot.connector.exchange.injective_v2.injective_v2_api_order_book_data_source import (
    InjectiveV2APIOrderBookDataSource,
)
from hummingbot.connector.exchange.injective_v2.injective_v2_utils import InjectiveConfigMap
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder
from hummingbot.connector.gateway.gateway_order_tracker import GatewayOrderTracker
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.market_order import MarketOrder
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TradeFeeBase, TradeFeeSchema
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.event_forwarder import EventForwarder
from hummingbot.core.event.events import AccountEvent, BalanceUpdateEvent, MarketEvent
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class InjectiveV2Exchange(ExchangePyBase):
    web_utils = web_utils

    def __init__(
            self,
            connector_configuration: InjectiveConfigMap,
            balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
            rate_limits_share_pct: Decimal = Decimal("100"),
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            **kwargs,
    ):
        self._orders_processing_delta_time = 0.5

        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._data_source = connector_configuration.create_data_source()
        self._rate_limits = connector_configuration.network.rate_limits()

        super().__init__(balance_asset_limit, rate_limits_share_pct)
        self._data_source.configure_throttler(throttler=self._throttler)
        self._forwarders = []
        self._configure_event_forwarders()
        self._latest_polled_order_fill_time: float = self._time()
        self._orders_transactions_check_task: Optional[asyncio.Task] = None
        self._orders_queued_to_create: List[GatewayInFlightOrder] = []
        self._orders_queued_to_cancel: List[GatewayInFlightOrder] = []

        self._orders_transactions_check_task = None
        self._queued_orders_task = None
        self._all_trading_events_queue = asyncio.Queue()

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def authenticator(self) -> AuthBase:
        return None

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return self._rate_limits

    @property
    def domain(self) -> str:
        return self._data_source.network_name

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self) -> str:
        raise NotImplementedError

    @property
    def trading_pairs_request_path(self) -> str:
        raise NotImplementedError

    @property
    def check_network_request_path(self) -> str:
        raise NotImplementedError

    @property
    def trading_pairs(self) -> List[str]:
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return False

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def status_dict(self) -> Dict[str, bool]:
        status = super().status_dict
        status["data_source_initialized"] = self._data_source.is_started()
        return status

    async def start_network(self):
        await super().start_network()

        market_ids = [
            await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            for trading_pair in self._trading_pairs
        ]
        await self._data_source.start(market_ids=market_ids)

        if self.is_trading_required:
            self._orders_transactions_check_task = safe_ensure_future(self._check_orders_transactions())
            self._queued_orders_task = safe_ensure_future(self._process_queued_orders())

    async def stop_network(self):
        """
        This function is executed when the connector is stopped. It performs a general cleanup and stops all background
        tasks that require the connection with the exchange to work.
        """
        await super().stop_network()
        await self._data_source.stop()
        self._forwarders = []
        if self._orders_transactions_check_task is not None:
            self._orders_transactions_check_task.cancel()
            self._orders_transactions_check_task = None
        if self._queued_orders_task is not None:
            self._queued_orders_task.cancel()
            self._queued_orders_task = None

    def supported_order_types(self) -> List[OrderType]:
        return self._data_source.supported_order_types()

    def start_tracking_order(
        self,
        order_id: str,
        exchange_order_id: Optional[str],
        trading_pair: str,
        trade_type: TradeType,
        price: Decimal,
        amount: Decimal,
        order_type: OrderType,
        **kwargs,
    ):
        self._order_tracker.start_tracking_order(
            GatewayInFlightOrder(
                client_order_id=order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=trading_pair,
                order_type=order_type,
                trade_type=trade_type,
                amount=amount,
                price=price,
                creation_timestamp=self.current_timestamp,
            )
        )

    def batch_order_create(self, orders_to_create: List[Union[MarketOrder, LimitOrder]]) -> List[LimitOrder]:
        """
        Issues a batch order creation as a single API request for exchanges that implement this feature. The default
        implementation of this method is to send the requests discretely (one by one).
        :param orders_to_create: A list of LimitOrder or MarketOrder objects representing the orders to create. The order IDs
            can be blanc.
        :returns: A tuple composed of LimitOrder or MarketOrder objects representing the created orders, complete with the generated
            order IDs.
        """
        orders_with_ids_to_create = []
        for order in orders_to_create:
            client_order_id = get_new_client_order_id(
                is_buy=order.is_buy,
                trading_pair=order.trading_pair,
                hbot_order_id_prefix=self.client_order_id_prefix,
                max_id_len=self.client_order_id_max_length,
            )
            orders_with_ids_to_create.append(order.copy_with_id(client_order_id=client_order_id))
        safe_ensure_future(self._execute_batch_order_create(orders_to_create=orders_with_ids_to_create))
        return orders_with_ids_to_create

    def batch_order_cancel(self, orders_to_cancel: List[LimitOrder]):
        """
        Issues a batch order cancelation as a single API request for exchanges that implement this feature. The default
        implementation of this method is to send the requests discretely (one by one).
        :param orders_to_cancel: A list of the orders to cancel.
        """
        safe_ensure_future(coro=self._execute_batch_cancel(orders_to_cancel=orders_to_cancel))

    async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
        """
        Cancels all currently active orders. The cancellations are performed in parallel tasks.

        :param timeout_seconds: the maximum time (in seconds) the cancel logic should run

        :return: a list of CancellationResult instances, one for each of the orders to be cancelled
        """
        incomplete_orders = {}
        limit_orders = []
        successful_cancellations = []

        for order in self.in_flight_orders.values():
            if not order.is_done:
                incomplete_orders[order.client_order_id] = order
                limit_orders.append(order.to_limit_order())

        if len(limit_orders) > 0:
            try:
                async with timeout(timeout_seconds):
                    cancellation_results = await self._execute_batch_cancel(orders_to_cancel=limit_orders)
                    for cr in cancellation_results:
                        if cr.success:
                            del incomplete_orders[cr.order_id]
                            successful_cancellations.append(CancellationResult(cr.order_id, True))
            except Exception:
                self.logger().network(
                    "Unexpected error cancelling orders.",
                    exc_info=True,
                    app_warning_msg="Failed to cancel order. Check API key and network connection."
                )
        failed_cancellations = [CancellationResult(oid, False) for oid in incomplete_orders.keys()]
        return successful_cancellations + failed_cancellations

    async def cancel_all_subaccount_orders(self):
        markets_ids = [await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                       for trading_pair in self.trading_pairs]
        await self._data_source.cancel_all_subaccount_orders(spot_markets_ids=markets_ids)

    async def check_network(self) -> NetworkStatus:
        """
        Checks connectivity with the exchange using the API
        """
        try:
            status = await self._data_source.check_network()
        except asyncio.CancelledError:
            raise
        except Exception:
            status = NetworkStatus.NOT_CONNECTED
        return status

    def trigger_event(self, event_tag: Enum, message: any):
        # Reimplemented because Injective connector has trading pairs with modified token names, because market tickers
        # are not always unique.
        # We need to change the original trading pair in all events to the real tokens trading pairs to not impact the
        # bot events processing
        trading_pair = getattr(message, "trading_pair", None)
        if trading_pair is not None:
            new_trading_pair = self._data_source.real_tokens_spot_trading_pair(unique_trading_pair=trading_pair)
            if isinstance(message, tuple):
                message = message._replace(trading_pair=new_trading_pair)
            else:
                setattr(message, "trading_pair", new_trading_pair)

        super().trigger_event(event_tag=event_tag, message=message)

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception) -> bool:
        return False

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_FOUND_ERROR_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # For Injective the cancelation is done by sending a transaction to the chain.
        # The cancel request is not validated until the transaction is included in a block, and so this does not apply
        return False

    async def _place_cancel(self, order_id: str, tracked_order: GatewayInFlightOrder):
        # Not required because of _execute_order_cancel redefinition
        raise NotImplementedError

    async def _execute_order_cancel(self, order: GatewayInFlightOrder) -> str:
        # Order cancelation requests for single orders are queued to be executed in batch if possible
        self._orders_queued_to_cancel.append(order)
        return None

    async def _place_order(self, order_id: str, trading_pair: str, amount: Decimal, trade_type: TradeType,
                           order_type: OrderType, price: Decimal, **kwargs) -> Tuple[str, float]:
        # Not required because of _place_order_and_process_update redefinition
        raise NotImplementedError

    async def _create_order(self,
                            trade_type: TradeType,
                            order_id: str,
                            trading_pair: str,
                            amount: Decimal,
                            order_type: OrderType,
                            price: Optional[Decimal] = None,
                            **kwargs):
        """
        Creates an order in the exchange using the parameters to configure it

        :param trade_type: the side of the order (BUY of SELL)
        :param order_id: the id that should be assigned to the order (the client id)
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        """
        try:
            if price is None or price.is_nan():
                calculated_price = self.get_price_for_volume(
                    trading_pair=trading_pair,
                    is_buy=trade_type == TradeType.BUY,
                    volume=amount,
                ).result_price
            else:
                calculated_price = price

            calculated_price = self.quantize_order_price(trading_pair, calculated_price)

            await super()._create_order(
                trade_type=trade_type,
                order_id=order_id,
                trading_pair=trading_pair,
                amount=amount,
                order_type=order_type,
                price=calculated_price,
                ** kwargs
            )

        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self._on_order_failure(
                order_id=order_id,
                trading_pair=trading_pair,
                amount=amount,
                trade_type=trade_type,
                order_type=order_type,
                price=price,
                exception=ex,
                **kwargs,
            )

    async def _place_order_and_process_update(self, order: GatewayInFlightOrder, **kwargs) -> str:
        # Order creation requests for single orders are queued to be executed in batch if possible
        self._orders_queued_to_create.append(order)
        return None

    async def _execute_batch_order_create(self, orders_to_create: List[Union[MarketOrder, LimitOrder]]):
        inflight_orders_to_create = []
        for order in orders_to_create:
            valid_order = await self._start_tracking_and_validate_order(
                trade_type=TradeType.BUY if order.is_buy else TradeType.SELL,
                order_id=order.client_order_id,
                trading_pair=order.trading_pair,
                amount=order.quantity,
                order_type=order.order_type(),
                price=order.price,
            )
            if valid_order is not None:
                inflight_orders_to_create.append(valid_order)
        await self._execute_batch_inflight_order_create(inflight_orders_to_create=inflight_orders_to_create)

    async def _execute_batch_inflight_order_create(self, inflight_orders_to_create: List[GatewayInFlightOrder]):
        try:
            place_order_results = await self._data_source.create_orders(
                spot_orders=inflight_orders_to_create
            )
            for place_order_result, in_flight_order in (
                zip(place_order_results, inflight_orders_to_create)
            ):
                if place_order_result.exception:
                    self._on_order_creation_failure(
                        order_id=in_flight_order.client_order_id,
                        trading_pair=in_flight_order.trading_pair,
                        amount=in_flight_order.amount,
                        trade_type=in_flight_order.trade_type,
                        order_type=in_flight_order.order_type,
                        price=in_flight_order.price,
                        exception=place_order_result.exception,
                    )
                else:
                    self._update_order_after_creation_success(
                        exchange_order_id=place_order_result.exchange_order_id,
                        order=in_flight_order,
                        update_timestamp=self.current_timestamp,
                        misc_updates=place_order_result.misc_updates,
                    )
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self.logger().network("Batch order create failed.")
            for order in inflight_orders_to_create:
                self._on_order_creation_failure(
                    order_id=order.client_order_id,
                    trading_pair=order.trading_pair,
                    amount=order.amount,
                    trade_type=order.trade_type,
                    order_type=order.order_type,
                    price=order.price,
                    exception=ex,
                )

    async def _start_tracking_and_validate_order(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        order_type: OrderType,
        price: Optional[Decimal] = None,
        **kwargs
    ) -> Optional[GatewayInFlightOrder]:
        trading_rule = self._trading_rules[trading_pair]

        if price is None:
            calculated_price = self.get_price_for_volume(
                trading_pair=trading_pair,
                is_buy=trade_type == TradeType.BUY,
                volume=amount,
            ).result_price
            calculated_price = self.quantize_order_price(trading_pair, calculated_price)
        else:
            calculated_price = price

        price = self.quantize_order_price(trading_pair, calculated_price)
        amount = self.quantize_order_amount(trading_pair=trading_pair, amount=amount)

        self.start_tracking_order(
            order_id=order_id,
            exchange_order_id=None,
            trading_pair=trading_pair,
            order_type=order_type,
            trade_type=trade_type,
            price=price,
            amount=amount,
            **kwargs,
        )
        order = self._order_tracker.active_orders[order_id]

        if order_type not in self.supported_order_types():
            self.logger().error(f"{order_type} is not in the list of supported order types")
            self._update_order_after_creation_failure(order_id=order_id, trading_pair=trading_pair)
            order = None
        elif amount < trading_rule.min_order_size:
            self.logger().warning(f"{trade_type.name.title()} order amount {amount} is lower than the minimum order"
                                  f" size {trading_rule.min_order_size}. The order will not be created.")
            self._update_order_after_creation_failure(order_id=order_id, trading_pair=trading_pair)
            order = None
        elif price is not None and amount * price < trading_rule.min_notional_size:
            self.logger().warning(f"{trade_type.name.title()} order notional {amount * price} is lower than the "
                                  f"minimum notional size {trading_rule.min_notional_size}. "
                                  "The order will not be created.")
            self._update_order_after_creation_failure(order_id=order_id, trading_pair=trading_pair)
            order = None

        return order

    def _update_order_after_creation_success(
        self,
        exchange_order_id: Optional[str],
        order: GatewayInFlightOrder,
        update_timestamp: float,
        misc_updates: Optional[Dict[str, Any]] = None
    ):
        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order.client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=order.current_state,
            misc_updates=misc_updates,
        )
        self.logger().debug(f"\nCreated order {order.client_order_id} ({exchange_order_id}) with TX {misc_updates}")
        self._order_tracker.process_order_update(order_update)

    def _on_order_creation_failure(
        self,
        order_id: str,
        trading_pair: str,
        amount: Decimal,
        trade_type: TradeType,
        order_type: OrderType,
        price: Optional[Decimal],
        exception: Exception,
    ):
        self.logger().network(
            f"Error submitting {trade_type.name.lower()} {order_type.name.upper()} order to {self.name_cap} for "
            f"{amount} {trading_pair} {price}.",
            exc_info=exception,
            app_warning_msg=f"Failed to submit buy order to {self.name_cap}. Check API key and network connection."
        )
        self._update_order_after_creation_failure(order_id=order_id, trading_pair=trading_pair)

    def _update_order_after_creation_failure(self, order_id: str, trading_pair: str):
        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order_id,
            trading_pair=trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=OrderState.FAILED,
        )
        self._order_tracker.process_order_update(order_update)

    async def _execute_batch_cancel(self, orders_to_cancel: List[LimitOrder]) -> List[CancellationResult]:
        results = []
        tracked_orders_to_cancel = []

        for order in orders_to_cancel:
            tracked_order = self._order_tracker.all_updatable_orders.get(order.client_order_id)
            if tracked_order is not None:
                tracked_orders_to_cancel.append(tracked_order)
            else:
                results.append(CancellationResult(order_id=order.client_order_id, success=False))

        if len(tracked_orders_to_cancel) > 0:
            results.extend(await self._execute_batch_order_cancel(orders_to_cancel=tracked_orders_to_cancel))

        return results

    async def _execute_batch_order_cancel(self, orders_to_cancel: List[GatewayInFlightOrder]) -> List[CancellationResult]:
        try:
            cancel_order_results = await self._data_source.cancel_orders(spot_orders=orders_to_cancel)
            cancelation_results = []
            for cancel_order_result in cancel_order_results:
                success = True
                if cancel_order_result.not_found:
                    self.logger().warning(
                        f"Failed to cancel the order {cancel_order_result.client_order_id} due to the order"
                        f" not being found."
                    )
                    await self._order_tracker.process_order_not_found(
                        client_order_id=cancel_order_result.client_order_id
                    )
                    success = False
                elif cancel_order_result.exception is not None:
                    self.logger().error(
                        f"Failed to cancel order {cancel_order_result.client_order_id}",
                        exc_info=cancel_order_result.exception,
                    )
                    success = False
                else:
                    order_update: OrderUpdate = OrderUpdate(
                        client_order_id=cancel_order_result.client_order_id,
                        trading_pair=cancel_order_result.trading_pair,
                        update_timestamp=self.current_timestamp,
                        new_state=(OrderState.CANCELED
                                   if self.is_cancel_request_in_exchange_synchronous
                                   else OrderState.PENDING_CANCEL),
                        misc_updates=cancel_order_result.misc_updates,
                    )
                    self._order_tracker.process_order_update(order_update)
                cancelation_results.append(
                    CancellationResult(order_id=cancel_order_result.client_order_id, success=success)
                )
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Failed to cancel orders {', '.join([o.client_order_id for o in orders_to_cancel])}",
                exc_info=True,
            )
            cancelation_results = [
                CancellationResult(order_id=order.client_order_id, success=False)
                for order in orders_to_cancel
            ]

        return cancelation_results

    def _update_order_after_cancelation_success(self, order: GatewayInFlightOrder):
        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order.client_order_id,
            trading_pair=order.trading_pair,
            update_timestamp=self.current_timestamp,
            new_state=(OrderState.CANCELED
                       if self.is_cancel_request_in_exchange_synchronous
                       else OrderState.PENDING_CANCEL),
        )
        self._order_tracker.process_order_update(order_update)

    def _get_fee(self, base_currency: str, quote_currency: str, order_type: OrderType, order_side: TradeType,
                 amount: Decimal, price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)
        if trading_pair in self._trading_fees:
            fee_schema: TradeFeeSchema = self._trading_fees[trading_pair]
            fee_rate = fee_schema.maker_percent_fee_decimal if is_maker else fee_schema.taker_percent_fee_decimal
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=fee_schema,
                trade_type=order_side,
                percent=fee_rate,
                percent_token=fee_schema.percent_fee_token,
            )
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    async def _update_trading_fees(self):
        self._trading_fees = await self._data_source.get_spot_trading_fees()

    async def _user_stream_event_listener(self):
        while True:
            try:
                event_message = await self._all_trading_events_queue.get()
                channel = event_message["channel"]
                event_data = event_message["data"]

                if channel == "transaction":
                    transaction_hash = event_data["hash"]
                    await self._check_created_orders_status_for_transaction(transaction_hash=transaction_hash)
                elif channel == "trade":
                    trade_update = event_data
                    self._order_tracker.process_trade_update(trade_update)
                elif channel == "order":
                    order_update = event_data
                    tracked_order = self._order_tracker.all_updatable_orders.get(order_update.client_order_id)
                    if tracked_order is not None:
                        is_partial_fill = order_update.new_state == OrderState.FILLED and not tracked_order.is_filled
                        if not is_partial_fill:
                            self._order_tracker.process_order_update(order_update=order_update)
                elif channel == "order_failure":
                    original_order_update = event_data
                    tracked_order = self._order_tracker.all_updatable_orders.get(original_order_update.client_order_id)
                    if tracked_order is not None:
                        # we need to set the trading_pair in the order update because that info is not included in the chain stream update
                        order_update = OrderUpdate(
                            trading_pair=tracked_order.trading_pair,
                            update_timestamp=original_order_update.update_timestamp,
                            new_state=original_order_update.new_state,
                            client_order_id=original_order_update.client_order_id,
                            exchange_order_id=original_order_update.exchange_order_id,
                            misc_updates=original_order_update.misc_updates,
                        )
                        self._order_tracker.process_order_update(order_update=order_update)
                elif channel == "balance":
                    if event_data.total_balance is not None:
                        self._account_balances[event_data.asset_name] = event_data.total_balance
                    if event_data.available_balance is not None:
                        self._account_available_balances[event_data.asset_name] = event_data.available_balance

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error in user stream listener loop")

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        # Not used in Injective
        raise NotImplementedError  # pragma: no cover

    async def _update_trading_rules(self):
        await self._data_source.update_markets()
        await self._initialize_trading_pair_symbol_map()
        trading_rules_list = await self._data_source.spot_trading_rules()
        trading_rules = {}
        for trading_rule in trading_rules_list:
            trading_rules[trading_rule.trading_pair] = trading_rule
        self._trading_rules.clear()
        self._trading_rules.update(trading_rules)

    async def _update_balances(self):
        all_balances = await self._data_source.all_account_balances()

        self._account_available_balances.clear()
        self._account_balances.clear()

        for token, token_balance_info in all_balances.items():
            self._account_balances[token] = token_balance_info["total_balance"]
            self._account_available_balances[token] = token_balance_info["available_balance"]

    async def _all_trade_updates_for_order(self, order: GatewayInFlightOrder) -> List[TradeUpdate]:
        # Not required because of _update_orders_fills redefinition
        raise NotImplementedError

    async def _update_orders_fills(self, orders: List[GatewayInFlightOrder]):
        oldest_order_creation_time = self.current_timestamp
        all_market_ids = set()

        for order in orders:
            oldest_order_creation_time = min(oldest_order_creation_time, order.creation_timestamp)
            all_market_ids.add(await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair))

        try:
            start_time = min(oldest_order_creation_time, self._latest_polled_order_fill_time)
            trade_updates = await self._data_source.spot_trade_updates(market_ids=all_market_ids, start_time=start_time)
            for trade_update in trade_updates:
                self._latest_polled_order_fill_time = max(self._latest_polled_order_fill_time, trade_update.fill_timestamp)
                self._order_tracker.process_trade_update(trade_update)
        except asyncio.CancelledError:
            raise
        except Exception as ex:
            self.logger().warning(
                f"Failed to fetch trade updates. Error: {ex}",
                exc_info=ex,
            )

    async def _request_order_status(self, tracked_order: GatewayInFlightOrder) -> OrderUpdate:
        # Not required due to the redefinition of _update_orders_with_error_handler
        raise NotImplementedError

    async def _update_orders_with_error_handler(self, orders: List[GatewayInFlightOrder], error_handler: Callable):
        oldest_order_creation_time = self.current_timestamp
        all_market_ids = set()
        orders_by_id = {}

        for order in orders:
            oldest_order_creation_time = min(oldest_order_creation_time, order.creation_timestamp)
            all_market_ids.add(await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair))
            orders_by_id[order.client_order_id] = order

        try:
            order_updates = await self._data_source.spot_order_updates(
                market_ids=all_market_ids,
                start_time=oldest_order_creation_time - self.LONG_POLL_INTERVAL
            )

            for order_update in order_updates:
                tracked_order = orders_by_id.get(order_update.client_order_id)
                if tracked_order is not None:
                    try:
                        if tracked_order.current_state == OrderState.PENDING_CREATE and order_update.new_state != OrderState.OPEN:
                            open_update = OrderUpdate(
                                trading_pair=order_update.trading_pair,
                                update_timestamp=order_update.update_timestamp,
                                new_state=OrderState.OPEN,
                                client_order_id=order_update.client_order_id,
                                exchange_order_id=order_update.exchange_order_id,
                                misc_updates=order_update.misc_updates,
                            )
                            self._order_tracker.process_order_update(open_update)

                        del orders_by_id[order_update.client_order_id]
                        self._order_tracker.process_order_update(order_update)
                    except asyncio.CancelledError:
                        raise
                    except Exception as ex:
                        await error_handler(tracked_order, ex)

            for order in orders_by_id.values():
                not_found_error = RuntimeError(
                    f"There was a problem updating order {order.client_order_id} "
                    f"({CONSTANTS.ORDER_NOT_FOUND_ERROR_MESSAGE})"
                )
                await error_handler(order, not_found_error)
        except asyncio.CancelledError:
            raise
        except Exception as request_error:
            for order in orders_by_id.values():
                await error_handler(order, request_error)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return WebAssistantsFactory(throttler=self._throttler)

    def _create_order_tracker(self) -> ClientOrderTracker:
        tracker = GatewayOrderTracker(connector=self)
        return tracker

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return InjectiveV2APIOrderBookDataSource(
            trading_pairs=self.trading_pairs,
            connector=self,
            data_source=self._data_source,
            domain=self.domain
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        # Not used in Injective
        raise NotImplementedError  # pragma: no cover

    def _is_user_stream_initialized(self):
        # Injective does not have private websocket endpoints
        return self._data_source.is_started()

    def _create_user_stream_tracker(self):
        # Injective does not use a tracker for the private streams
        return None

    def _create_user_stream_tracker_task(self):
        # Injective does not use a tracker for the private streams
        return None

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        # Not used in Injective
        raise NotImplementedError()  # pragma: no cover

    async def _initialize_trading_pair_symbol_map(self):
        exchange_info = None
        try:
            mapping = await self._data_source.spot_market_and_trading_pair_map()
            self._set_trading_pair_symbol_map(mapping)
        except Exception:
            self.logger().exception("There was an error requesting exchange info.")
        return exchange_info

    def _configure_event_forwarders(self):
        event_forwarder = EventForwarder(to_function=self._process_user_trade_update)
        self._forwarders.append(event_forwarder)
        self._data_source.add_listener(event_tag=MarketEvent.TradeUpdate, listener=event_forwarder)

        event_forwarder = EventForwarder(to_function=self._process_user_order_update)
        self._forwarders.append(event_forwarder)
        self._data_source.add_listener(event_tag=MarketEvent.OrderUpdate, listener=event_forwarder)

        event_forwarder = EventForwarder(to_function=self._process_user_order_failure_update)
        self._forwarders.append(event_forwarder)
        self._data_source.add_listener(event_tag=MarketEvent.OrderFailure, listener=event_forwarder)

        event_forwarder = EventForwarder(to_function=self._process_balance_event)
        self._forwarders.append(event_forwarder)
        self._data_source.add_listener(event_tag=AccountEvent.BalanceEvent, listener=event_forwarder)

        event_forwarder = EventForwarder(to_function=self._process_transaction_event)
        self._forwarders.append(event_forwarder)
        self._data_source.add_listener(event_tag=InjectiveEvent.ChainTransactionEvent, listener=event_forwarder)

    def _process_balance_event(self, event: BalanceUpdateEvent):
        self._all_trading_events_queue.put_nowait(
            {"channel": "balance", "data": event}
        )

    def _process_user_order_update(self, order_update: OrderUpdate):
        self._all_trading_events_queue.put_nowait(
            {"channel": "order", "data": order_update}
        )

    def _process_user_order_failure_update(self, order_update: OrderUpdate):
        self._all_trading_events_queue.put_nowait(
            {"channel": "order_failure", "data": order_update}
        )

    def _process_user_trade_update(self, trade_update: TradeUpdate):
        self._all_trading_events_queue.put_nowait(
            {"channel": "trade", "data": trade_update}
        )

    def _process_transaction_event(self, transaction_event: Dict[str, Any]):
        self._all_trading_events_queue.put_nowait(
            {"channel": "transaction", "data": transaction_event}
        )

    async def _check_orders_transactions(self):
        while True:
            try:
                # Executing the process shielded from this async task to isolate it from network disconnections
                # (network disconnections cancel this task)
                task = asyncio.create_task(self._check_orders_creation_transactions())
                await asyncio.shield(task)
                await self._sleep(CONSTANTS.TRANSACTIONS_CHECK_INTERVAL)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while running the transactions check process", exc_info=True)
                await self._sleep(0.5)

    async def _check_orders_creation_transactions(self):
        orders: List[GatewayInFlightOrder] = self._order_tracker.active_orders.values()
        orders_by_creation_tx = defaultdict(list)

        for order in orders:
            if order.creation_transaction_hash is not None and order.is_pending_create:
                orders_by_creation_tx[order.creation_transaction_hash].append(order)

        for transaction_hash, orders in orders_by_creation_tx.items():
            try:
                order_updates = await self._data_source.order_updates_for_transaction(
                    transaction_hash=transaction_hash, spot_orders=orders
                )
                for order_update in order_updates:
                    self._order_tracker.process_order_update(order_update=order_update)

            except ValueError:
                self.logger().debug(f"Transaction not included in a block yet ({transaction_hash})")

    async def _check_created_orders_status_for_transaction(self, transaction_hash: str):
        transaction_orders = []
        order: GatewayInFlightOrder
        for order in self.in_flight_orders.values():
            if order.creation_transaction_hash == transaction_hash and order.is_pending_create:
                transaction_orders.append(order)

        if len(transaction_orders) > 0:
            order_updates = await self._data_source.order_updates_for_transaction(
                transaction_hash=transaction_hash, spot_orders=transaction_orders
            )

            for order_update in order_updates:
                self._order_tracker.process_order_update(order_update=order_update)

    async def _process_queued_orders(self):
        while True:
            try:
                # Executing the batch cancelation and creation process shielded from this async task to isolate the
                # creation/cancelation process from network disconnections (network disconnections cancel this task)
                task = asyncio.create_task(self._cancel_and_create_queued_orders())
                await asyncio.shield(task)
                sleep_time = (self.clock.tick_size * 0.5
                              if self.clock is not None
                              else self._orders_processing_delta_time)
                await self._sleep(sleep_time)
            except NotImplementedError:
                raise
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().exception("Unexpected error while processing queued individual orders", exc_info=True)
                await self._sleep(self.clock.tick_size * 0.5)

    async def _cancel_and_create_queued_orders(self):
        if len(self._orders_queued_to_cancel) > 0:
            orders = [order.to_limit_order() for order in self._orders_queued_to_cancel]
            self._orders_queued_to_cancel = []
            await self._execute_batch_cancel(orders_to_cancel=orders)
        if len(self._orders_queued_to_create) > 0:
            orders = self._orders_queued_to_create
            self._orders_queued_to_create = []
            await self._execute_batch_inflight_order_create(inflight_orders_to_create=orders)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        market_id = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        last_price = await self._data_source.last_traded_price(market_id=market_id)
        return float(last_price)

    def _get_poll_interval(self, timestamp: float) -> float:
        last_recv_diff = timestamp - self._data_source.last_received_message_timestamp
        poll_interval = (
            self.SHORT_POLL_INTERVAL
            if last_recv_diff > self.TICK_INTERVAL_LIMIT
            else self.LONG_POLL_INTERVAL
        )
        return poll_interval



# distutils: language=c++

from hummingbot.core.time_iterator cimport TimeIterator
from hummingbot.core.event.event_listener cimport EventListener

from .order_tracker cimport OrderTracker

cdef class StrategyBase(TimeIterator):
    cdef:
        set _sb_markets
        EventListener _sb_create_buy_order_listener
        EventListener _sb_create_sell_order_listener
        EventListener _sb_fill_order_listener
        EventListener _sb_fail_order_listener
        EventListener _sb_cancel_order_listener
        EventListener _sb_expire_order_listener
        EventListener _sb_complete_buy_order_listener
        EventListener _sb_complete_sell_order_listener
        EventListener _sb_complete_funding_payment_listener
        EventListener _sb_position_mode_change_success_listener
        EventListener _sb_position_mode_change_failure_listener
        EventListener _sb_range_position_liquidity_added_listener
        EventListener _sb_range_position_liquidity_removed_listener
        EventListener _sb_range_position_update_listener
        EventListener _sb_range_position_update_failure_listener
        EventListener _sb_range_position_fee_collected_listener
        EventListener _sb_range_position_closed_listener
        bint _sb_delegate_lock
        public OrderTracker _sb_order_tracker

    cdef c_add_markets(self, list markets)
    cdef c_remove_markets(self, list markets)
    cdef c_did_create_buy_order(self, object order_created_event)
    cdef c_did_create_sell_order(self, object order_created_event)
    cdef c_did_fill_order(self, object order_filled_event)
    cdef c_did_fail_order(self, object order_failed_event)
    cdef c_did_cancel_order(self, object cancelled_event)
    cdef c_did_expire_order(self, object expired_event)
    cdef c_did_complete_buy_order(self, object order_completed_event)
    cdef c_did_complete_sell_order(self, object order_completed_event)
    cdef c_did_complete_funding_payment(self, object funding_payment_completed_event)
    cdef c_did_change_position_mode_succeed(self, object position_mode_changed_event)
    cdef c_did_change_position_mode_fail(self, object position_mode_changed_event)
    cdef c_did_add_liquidity(self, object add_liquidity_event)
    cdef c_did_remove_liquidity(self, object remove_liquidity_event)
    cdef c_did_fail_lp_update(self, object fail_lp_update_event)

    cdef c_did_fail_order_tracker(self, object order_failed_event)
    cdef c_did_cancel_order_tracker(self, object order_cancelled_event)
    cdef c_did_expire_order_tracker(self, object order_expired_event)
    cdef c_did_complete_buy_order_tracker(self, object order_completed_event)
    cdef c_did_complete_sell_order_tracker(self, object order_completed_event)

    cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
                                        object price = *, double expiration_seconds = *, position_action = *)
    cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
                                         object price = *, double expiration_seconds = *, position_action = *, )
    cdef c_cancel_order(self, object market_pair, str order_id)

    cdef c_start_tracking_limit_order(self, object market_pair, str order_id, bint is_buy, object price,
                                      object quantity)
    cdef c_stop_tracking_limit_order(self, object market_pair, str order_id)
    cdef c_start_tracking_market_order(self, object market_pair, str order_id, bint is_buy, object quantity)
    cdef c_stop_tracking_market_order(self, object market_pair, str order_id)
    cdef c_track_restored_orders(self, object market_pair)
    cdef object c_sum_flat_fees(self,
                                str quote_currency,
                                list flat_fees)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PrivateOrdersV3Api(_message.Message):
    __slots__ = ("id", "clientId", "price", "quantity", "amount", "avgPrice", "orderType", "tradeType", "isMaker", "remainAmount", "remainQuantity", "lastDealQuantity", "cumulativeQuantity", "cumulativeAmount", "status", "createTime", "market", "triggerType", "triggerPrice", "state", "ocoId", "routeFactor", "symbolId", "marketId", "marketCurrencyId", "currencyId")
    ID_FIELD_NUMBER: _ClassVar[int]
    CLIENTID_FIELD_NUMBER: _ClassVar[int]
    PRICE_FIELD_NUMBER: _ClassVar[int]
    QUANTITY_FIELD_NUMBER: _ClassVar[int]
    AMOUNT_FIELD_NUMBER: _ClassVar[int]
    AVGPRICE_FIELD_NUMBER: _ClassVar[int]
    ORDERTYPE_FIELD_NUMBER: _ClassVar[int]
    TRADETYPE_FIELD_NUMBER: _ClassVar[int]
    ISMAKER_FIELD_NUMBER: _ClassVar[int]
    REMAINAMOUNT_FIELD_NUMBER: _ClassVar[int]
    REMAINQUANTITY_FIELD_NUMBER: _ClassVar[int]
    LASTDEALQUANTITY_FIELD_NUMBER: _ClassVar[int]
    CUMULATIVEQUANTITY_FIELD_NUMBER: _ClassVar[int]
    CUMULATIVEAMOUNT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    CREATETIME_FIELD_NUMBER: _ClassVar[int]
    MARKET_FIELD_NUMBER: _ClassVar[int]
    TRIGGERTYPE_FIELD_NUMBER: _ClassVar[int]
    TRIGGERPRICE_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    OCOID_FIELD_NUMBER: _ClassVar[int]
    ROUTEFACTOR_FIELD_NUMBER: _ClassVar[int]
    SYMBOLID_FIELD_NUMBER: _ClassVar[int]
    MARKETID_FIELD_NUMBER: _ClassVar[int]
    MARKETCURRENCYID_FIELD_NUMBER: _ClassVar[int]
    CURRENCYID_FIELD_NUMBER: _ClassVar[int]
    id: str
    clientId: str
    price: str
    quantity: str
    amount: str
    avgPrice: str
    orderType: int
    tradeType: int
    isMaker: bool
    remainAmount: str
    remainQuantity: str
    lastDealQuantity: str
    cumulativeQuantity: str
    cumulativeAmount: str
    status: int
    createTime: int
    market: str
    triggerType: int
    triggerPrice: str
    state: int
    ocoId: str
    routeFactor: str
    symbolId: str
    marketId: str
    marketCurrencyId: str
    currencyId: str
    def __init__(self, id: _Optional[str] = ..., clientId: _Optional[str] = ..., price: _Optional[str] = ..., quantity: _Optional[str] = ..., amount: _Optional[str] = ..., avgPrice: _Optional[str] = ..., orderType: _Optional[int] = ..., tradeType: _Optional[int] = ..., isMaker: bool = ..., remainAmount: _Optional[str] = ..., remainQuantity: _Optional[str] = ..., lastDealQuantity: _Optional[str] = ..., cumulativeQuantity: _Optional[str] = ..., cumulativeAmount: _Optional[str] = ..., status: _Optional[int] = ..., createTime: _Optional[int] = ..., market: _Optional[str] = ..., triggerType: _Optional[int] = ..., triggerPrice: _Optional[str] = ..., state: _Optional[int] = ..., ocoId: _Optional[str] = ..., routeFactor: _Optional[str] = ..., symbolId: _Optional[str] = ..., marketId: _Optional[str] = ..., marketCurrencyId: _Optional[str] = ..., currencyId: _Optional[str] = ...) -> None: ...



import asyncio
import logging
from decimal import Decimal
from enum import Enum
from typing import Dict, List, Tuple

import pandas as pd

from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.connector.derivative.position import Position
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, TradeType
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.market_order import MarketOrder
from hummingbot.core.data_type.order_candidate import OrderCandidate, PerpetualOrderCandidate
from hummingbot.core.event.events import BuyOrderCompletedEvent, PositionModeChangeEvent, SellOrderCompletedEvent
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.spot_perpetual_arbitrage.arb_proposal import ArbProposal, ArbProposalSide
from hummingbot.strategy.strategy_py_base import StrategyPyBase

NaN = float("nan")
s_decimal_zero = Decimal(0)
spa_logger = None


class StrategyState(Enum):
    Closed = 0
    Opening = 1
    Opened = 2
    Closing = 3


class SpotPerpetualArbitrageStrategy(StrategyPyBase):
    """
    This strategy arbitrages between a spot and a perpetual exchange.
    For a given order amount, the strategy checks for price discrepancy between buy and sell price on the 2 exchanges.
    Since perpetual contract requires closing position before profit is realised, there are 2 stages to this arbitrage
    operation - first to open and second to close.
    """

    @classmethod
    def logger(cls) -> HummingbotLogger:
        global spa_logger
        if spa_logger is None:
            spa_logger = logging.getLogger(__name__)
        return spa_logger

    def init_params(self,
                    spot_market_info: MarketTradingPairTuple,
                    perp_market_info: MarketTradingPairTuple,
                    order_amount: Decimal,
                    perp_leverage: int,
                    min_opening_arbitrage_pct: Decimal,
                    min_closing_arbitrage_pct: Decimal,
                    spot_market_slippage_buffer: Decimal = Decimal("0"),
                    perp_market_slippage_buffer: Decimal = Decimal("0"),
                    next_arbitrage_opening_delay: float = 120,
                    status_report_interval: float = 10):
        """
        :param spot_market_info: The spot market info
        :param perp_market_info: The perpetual market info
        :param order_amount: The order amount
        :param perp_leverage: The leverage level to use on perpetual market
        :param min_opening_arbitrage_pct: The minimum spread to open arbitrage position (e.g. 0.0003 for 0.3%)
        :param min_closing_arbitrage_pct: The minimum spread to close arbitrage position (e.g. 0.0003 for 0.3%)
        :param spot_market_slippage_buffer: The buffer for which to adjust order price for higher chance of
        the order getting filled on spot market.
        :param perp_market_slippage_buffer: The slipper buffer for perpetual market.
        :param next_arbitrage_opening_delay: The number of seconds to delay before the next arb position can be opened
        :param status_report_interval: Amount of seconds to wait to refresh the status report
        """
        self._spot_market_info = spot_market_info
        self._perp_market_info = perp_market_info
        self._min_opening_arbitrage_pct = min_opening_arbitrage_pct
        self._min_closing_arbitrage_pct = min_closing_arbitrage_pct
        self._order_amount = order_amount
        self._perp_leverage = perp_leverage
        self._spot_market_slippage_buffer = spot_market_slippage_buffer
        self._perp_market_slippage_buffer = perp_market_slippage_buffer
        self._next_arbitrage_opening_delay = next_arbitrage_opening_delay
        self._next_arbitrage_opening_ts = 0  # next arbitrage opening timestamp
        self._all_markets_ready = False
        self._ev_loop = asyncio.get_event_loop()
        self._last_timestamp = 0
        self._status_report_interval = status_report_interval
        self.add_markets([spot_market_info.market, perp_market_info.market])

        self._main_task = None
        self._in_flight_opening_order_ids = []
        self._completed_opening_order_ids = []
        self._completed_closing_order_ids = []
        self._strategy_state = StrategyState.Closed
        self._ready_to_start = False
        self._last_arb_op_reported_ts = 0
        self._position_mode_ready = False
        self._position_mode_not_ready_counter = 0
        self._trading_started = False

    def all_markets_ready(self):
        return all([market.ready for market in self.active_markets])

    @property
    def strategy_state(self) -> StrategyState:
        return self._strategy_state

    @property
    def min_opening_arbitrage_pct(self) -> Decimal:
        return self._min_opening_arbitrage_pct

    @property
    def min_closing_arbitrage_pct(self) -> Decimal:
        return self._min_closing_arbitrage_pct

    @property
    def order_amount(self) -> Decimal:
        return self._order_amount

    @order_amount.setter
    def order_amount(self, value):
        self._order_amount = value

    @property
    def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
        return self._sb_order_tracker.market_pair_to_active_orders

    @property
    def perp_positions(self) -> List[Position]:
        return [s for s in self._perp_market_info.market.account_positions.values() if
                s.trading_pair == self._perp_market_info.trading_pair and s.amount != s_decimal_zero]

    def apply_initial_settings(self):
        self._perp_market_info.market.set_leverage(self._perp_market_info.trading_pair, self._perp_leverage)
        self._perp_market_info.market.set_position_mode(PositionMode.ONEWAY)

    def tick(self, timestamp: float):
        """
        Clock tick entry point, is run every second (on normal tick setting).
        :param timestamp: current tick timestamp
        """
        if not self._all_markets_ready or not self._position_mode_ready or not self._trading_started:
            self._all_markets_ready = self.all_markets_ready()
            if not self._all_markets_ready:
                return
            else:
                self.logger().info("Markets are ready.")

            if not self._position_mode_ready:
                self._position_mode_not_ready_counter += 1
                # Attempt to switch position mode every 10 ticks only to not to spam and DDOS
                if self._position_mode_not_ready_counter == 10:
                    self._perp_market_info.market.set_position_mode(PositionMode.ONEWAY)
                    self._position_mode_not_ready_counter = 0
                return
            self._position_mode_not_ready_counter = 0

            self.logger().info("Trading started.")
            self._trading_started = True

            if not self.check_budget_available():
                self.logger().info("Trading not possible.")
                return

            if self._perp_market_info.market.position_mode != PositionMode.ONEWAY or \
                    len(self.perp_positions) > 1:
                self.logger().info("This strategy supports only Oneway position mode. Attempting to switch ...")
                self._perp_market_info.market.set_position_mode(PositionMode.ONEWAY)
                return

            if len(self.perp_positions) == 1:
                adj_perp_amount = self._perp_market_info.market.quantize_order_amount(
                    self._perp_market_info.trading_pair, self._order_amount)
                if abs(self.perp_positions[0].amount) == adj_perp_amount:
                    self.logger().info(f"There is an existing {self._perp_market_info.trading_pair} "
                                       f"{self.perp_positions[0].position_side.name} position. The bot resumes "
                                       f"operation to close out the arbitrage position")
                    self._strategy_state = StrategyState.Opened
                    self._ready_to_start = True
                else:
                    self.logger().info(f"There is an existing {self._perp_market_info.trading_pair} "
                                       f"{self.perp_positions[0].position_side.name} position with unmatched "
                                       f"position amount. Please manually close out the position before starting "
                                       f"this strategy.")
                    return
            else:
                self._ready_to_start = True

        if self._ready_to_start and (self._main_task is None or self._main_task.done()):
            self._main_task = safe_ensure_future(self.main(timestamp))

    async def main(self, timestamp):
        """
        The main procedure for the arbitrage strategy.
        """
        self.update_strategy_state()
        if self._strategy_state in (StrategyState.Opening, StrategyState.Closing):
            return
        if self.strategy_state == StrategyState.Closed and self._next_arbitrage_opening_ts > self.current_timestamp:
            return
        proposals = await self.create_base_proposals()
        if self._strategy_state == StrategyState.Opened:
            perp_is_buy = False if self.perp_positions[0].amount > 0 else True
            proposals = [p for p in proposals if p.perp_side.is_buy == perp_is_buy and p.profit_pct() >=
                         self._min_closing_arbitrage_pct]
        else:
            proposals = [p for p in proposals if p.profit_pct() >= self._min_opening_arbitrage_pct]
        if len(proposals) == 0:
            return
        proposal = proposals[0]
        if self._last_arb_op_reported_ts + 60 < self.current_timestamp:
            pos_txt = "closing" if self._strategy_state == StrategyState.Opened else "opening"
            self.logger().info(f"Arbitrage position {pos_txt} opportunity found.")
            self.logger().info(f"Profitability ({proposal.profit_pct():.2%}) is now above min_{pos_txt}_arbitrage_pct.")
            self._last_arb_op_reported_ts = self.current_timestamp
        self.apply_slippage_buffers(proposal)
        if self.check_budget_constraint(proposal):
            self.execute_arb_proposal(proposal)

    def update_strategy_state(self):
        """
        Updates strategy state to either Opened or Closed if the condition is right.
        """
        if self._strategy_state == StrategyState.Opening and len(self._completed_opening_order_ids) == 2 and \
                self.perp_positions:
            self._strategy_state = StrategyState.Opened
            self._completed_opening_order_ids.clear()
        elif self._strategy_state == StrategyState.Closing and len(self._completed_closing_order_ids) == 2 and \
                len(self.perp_positions) == 0:
            self._strategy_state = StrategyState.Closed
            self._completed_closing_order_ids.clear()
            self._next_arbitrage_opening_ts = self.current_timestamp + self._next_arbitrage_opening_delay

    async def create_base_proposals(self) -> List[ArbProposal]:
        """
        Creates a list of 2 base proposals, no filter.
        :return: A list of 2 base proposals.
        """
        tasks = [self._spot_market_info.market.get_order_price(self._spot_market_info.trading_pair, True,
                                                               self._order_amount),
                 self._spot_market_info.market.get_order_price(self._spot_market_info.trading_pair, False,
                                                               self._order_amount),
                 self._perp_market_info.market.get_order_price(self._perp_market_info.trading_pair, True,
                                                               self._order_amount),
                 self._perp_market_info.market.get_order_price(self._perp_market_info.trading_pair, False,
                                                               self._order_amount)]
        prices = await safe_gather(*tasks, return_exceptions=True)
        spot_buy, spot_sell, perp_buy, perp_sell = [*prices]
        return [
            ArbProposal(ArbProposalSide(self._spot_market_info, True, spot_buy),
                        ArbProposalSide(self._perp_market_info, False, perp_sell),
                        self._order_amount),
            ArbProposal(ArbProposalSide(self._spot_market_info, False, spot_sell),
                        ArbProposalSide(self._perp_market_info, True, perp_buy),
                        self._order_amount)
        ]

    def apply_slippage_buffers(self, proposal: ArbProposal):
        """
        Updates arb_proposals by adjusting order price for slipper buffer percentage.
        E.g. if it is a buy order, for an order price of 100 and 1% slipper buffer, the new order price is 101,
        for a sell order, the new order price is 99.
        :param proposal: the arbitrage proposal
        """
        for arb_side in (proposal.spot_side, proposal.perp_side):
            market = arb_side.market_info.market
            # arb_side.amount = market.quantize_order_amount(arb_side.market_info.trading_pair, arb_side.amount)
            s_buffer = self._spot_market_slippage_buffer if market == self._spot_market_info.market \
                else self._perp_market_slippage_buffer
            if not arb_side.is_buy:
                s_buffer *= Decimal("-1")
            arb_side.order_price *= Decimal("1") + s_buffer
            arb_side.order_price = market.quantize_order_price(arb_side.market_info.trading_pair,
                                                               arb_side.order_price)

    def check_budget_available(self) -> bool:
        """
        Checks if there's any balance for trading to be possible at all
        :return: True if user has available balance enough for orders submission.
        """

        spot_base, spot_quote = self._spot_market_info.trading_pair.split("-")
        perp_base, perp_quote = self._perp_market_info.trading_pair.split("-")

        balance_spot_base = self._spot_market_info.market.get_available_balance(spot_base)
        balance_spot_quote = self._spot_market_info.market.get_available_balance(spot_quote)

        balance_perp_quote = self._perp_market_info.market.get_available_balance(perp_quote)

        if balance_spot_base == s_decimal_zero and balance_spot_quote == s_decimal_zero:
            self.logger().info(f"Cannot arbitrage, {self._spot_market_info.market.display_name} {spot_base} balance "
                               f"({balance_spot_base}) is 0 and {self._spot_market_info.market.display_name} {spot_quote} balance "
                               f"({balance_spot_quote}) is 0.")
            return False

        if balance_perp_quote == s_decimal_zero:
            self.logger().info(f"Cannot arbitrage, {self._perp_market_info.market.display_name} {perp_quote} balance "
                               f"({balance_perp_quote}) is 0.")
            return False

        return True

    def check_budget_constraint(self, proposal: ArbProposal) -> bool:
        """
        Check balances on both exchanges if there is enough to submit both orders in a proposal.
        :param proposal: An arbitrage proposal
        :return: True if user has available balance enough for both orders submission.
        """
        return self.check_spot_budget_constraint(proposal) and self.check_perpetual_budget_constraint(proposal)

    def check_spot_budget_constraint(self, proposal: ArbProposal) -> bool:
        """
        Check balance on spot exchange.
        :param proposal: An arbitrage proposal
        :return: True if user has available balance enough for both orders submission.
        """
        proposal_side = proposal.spot_side
        order_amount = proposal.order_amount
        market_info = proposal_side.market_info
        budget_checker = market_info.market.budget_checker
        order_candidate = OrderCandidate(
            trading_pair=market_info.trading_pair,
            is_maker=False,
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY if proposal_side.is_buy else TradeType.SELL,
            amount=order_amount,
            price=proposal_side.order_price,
        )

        adjusted_candidate_order = budget_checker.adjust_candidate(order_candidate, all_or_none=True)

        if adjusted_candidate_order.amount < order_amount:
            self.logger().info(
                f"Cannot arbitrage, {proposal_side.market_info.market.display_name} balance"
                f" is insufficient to place the order candidate {order_candidate}."
            )
            return False

        return True

    def check_perpetual_budget_constraint(self, proposal: ArbProposal) -> bool:
        """
        Check balance on spot exchange.
        :param proposal: An arbitrage proposal
        :return: True if user has available balance enough for both orders submission.
        """
        proposal_side = proposal.perp_side
        order_amount = proposal.order_amount
        market_info = proposal_side.market_info
        budget_checker = market_info.market.budget_checker

        position_close = False
        if self.perp_positions and abs(self.perp_positions[0].amount) == order_amount:
            perp_side = proposal.perp_side
            cur_perp_pos_is_buy = True if self.perp_positions[0].amount > 0 else False
            if perp_side != cur_perp_pos_is_buy:
                position_close = True

        order_candidate = PerpetualOrderCandidate(
            trading_pair=market_info.trading_pair,
            is_maker=False,
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY if proposal_side.is_buy else TradeType.SELL,
            amount=order_amount,
            price=proposal_side.order_price,
            leverage=Decimal(self._perp_leverage),
            position_close=position_close,
        )

        adjusted_candidate_order = budget_checker.adjust_candidate(order_candidate, all_or_none=True)

        if adjusted_candidate_order.amount < order_amount:
            self.logger().info(
                f"Cannot arbitrage, {proposal_side.market_info.market.display_name} balance"
                f" is insufficient to place the order candidate {order_candidate}."
            )
            return False

        return True

    def execute_arb_proposal(self, proposal: ArbProposal):
        """
        Execute both sides of the arbitrage trades concurrently.
        :param proposal: the arbitrage proposal
        """
        if proposal.order_amount == s_decimal_zero:
            return
        spot_side = proposal.spot_side
        spot_order_fn = self.buy_with_specific_market if spot_side.is_buy else self.sell_with_specific_market
        side = "BUY" if spot_side.is_buy else "SELL"
        self.log_with_clock(
            logging.INFO,
            f"Placing {side} order for {proposal.order_amount} {spot_side.market_info.base_asset} "
            f"at {spot_side.market_info.market.display_name} at {spot_side.order_price} price"
        )
        spot_order_fn(
            spot_side.market_info,
            proposal.order_amount,
            spot_side.market_info.market.get_taker_order_type(),
            spot_side.order_price,
        )
        perp_side = proposal.perp_side
        perp_order_fn = self.buy_with_specific_market if perp_side.is_buy else self.sell_with_specific_market
        side = "BUY" if perp_side.is_buy else "SELL"
        position_action = PositionAction.CLOSE if self._strategy_state == StrategyState.Opened else PositionAction.OPEN
        self.log_with_clock(
            logging.INFO,
            f"Placing {side} order for {proposal.order_amount} {perp_side.market_info.base_asset} "
            f"at {perp_side.market_info.market.display_name} at {perp_side.order_price} price to "
            f"{position_action.name} position."
        )
        perp_order_fn(
            perp_side.market_info,
            proposal.order_amount,
            perp_side.market_info.market.get_taker_order_type(),
            perp_side.order_price,
            position_action=position_action
        )
        if self._strategy_state == StrategyState.Opened:
            self._strategy_state = StrategyState.Closing
            self._completed_closing_order_ids.clear()
        else:
            self._strategy_state = StrategyState.Opening
            self._completed_opening_order_ids.clear()

    def active_positions_df(self) -> pd.DataFrame:
        """
        Returns a new dataframe on current active perpetual positions.
        """
        columns = ["Symbol", "Type", "Entry Price", "Amount", "Leverage", "Unrealized PnL"]
        data = []
        for pos in self.perp_positions:
            data.append([
                pos.trading_pair,
                "LONG" if pos.amount > 0 else "SHORT",
                pos.entry_price,
                pos.amount,
                pos.leverage,
                pos.unrealized_pnl
            ])

        return pd.DataFrame(data=data, columns=columns)

    async def format_status(self) -> str:
        """
        Returns a status string formatted to display nicely on terminal. The strings composes of 4 parts: markets,
        assets, spread and warnings(if any).
        """
        columns = ["Exchange", "Market", "Sell Price", "Buy Price", "Mid Price"]
        data = []
        for market_info in [self._spot_market_info, self._perp_market_info]:
            market, trading_pair, base_asset, quote_asset = market_info
            buy_price = await market.get_quote_price(trading_pair, True, self._order_amount)
            sell_price = await market.get_quote_price(trading_pair, False, self._order_amount)
            mid_price = (buy_price + sell_price) / 2
            data.append([
                market.display_name,
                trading_pair,
                float(sell_price),
                float(buy_price),
                float(mid_price)
            ])
        markets_df = pd.DataFrame(data=data, columns=columns)
        lines = []
        lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])

        # See if there're any active positions.
        if len(self.perp_positions) > 0:
            df = self.active_positions_df()
            lines.extend(["", "  Positions:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
        else:
            lines.extend(["", "  No active positions."])

        assets_df = self.wallet_balance_data_frame([self._spot_market_info, self._perp_market_info])
        lines.extend(["", "  Assets:"] +
                     ["    " + line for line in str(assets_df).split("\n")])

        proposals = await self.create_base_proposals()
        lines.extend(["", "  Opportunity:"] + self.short_proposal_msg(proposals))

        warning_lines = self.network_warning([self._spot_market_info])
        warning_lines.extend(self.network_warning([self._perp_market_info]))
        warning_lines.extend(self.balance_warning([self._spot_market_info]))
        warning_lines.extend(self.balance_warning([self._perp_market_info]))
        if len(warning_lines) > 0:
            lines.extend(["", "*** WARNINGS ***"] + warning_lines)

        return "\n".join(lines)

    def short_proposal_msg(self, arb_proposal: List[ArbProposal], indented: bool = True) -> List[str]:
        """
        Composes a short proposal message.
        :param arb_proposal: The arbitrage proposal
        :param indented: If the message should be indented (by 4 spaces)
        :return A list of messages
        """
        lines = []
        for proposal in arb_proposal:
            spot_side = "buy" if proposal.spot_side.is_buy else "sell"
            perp_side = "buy" if proposal.perp_side.is_buy else "sell"
            profit_pct = proposal.profit_pct()
            lines.append(f"{'    ' if indented else ''}{spot_side} at "
                         f"{proposal.spot_side.market_info.market.display_name}"
                         f", {perp_side} at {proposal.perp_side.market_info.market.display_name}: "
                         f"{profit_pct:.2%}")
        return lines

    @property
    def tracked_market_orders(self) -> List[Tuple[ConnectorBase, MarketOrder]]:
        return self._sb_order_tracker.tracked_market_orders

    @property
    def tracked_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        return self._sb_order_tracker.tracked_limit_orders

    def start(self, clock: Clock, timestamp: float):
        self._ready_to_start = False
        self.apply_initial_settings()

    def stop(self, clock: Clock):
        if self._main_task is not None:
            self._main_task.cancel()
            self._main_task = None
        self._ready_to_start = False

    def did_complete_buy_order(self, event: BuyOrderCompletedEvent):
        self.update_complete_order_id_lists(event.order_id)

    def did_complete_sell_order(self, event: SellOrderCompletedEvent):
        self.update_complete_order_id_lists(event.order_id)

    def did_change_position_mode_succeed(self, position_mode_changed_event: PositionModeChangeEvent):
        if position_mode_changed_event.position_mode is PositionMode.ONEWAY:
            self.logger().info(
                f"Changing position mode to {PositionMode.ONEWAY.name} succeeded.")
            self._position_mode_ready = True
        else:
            self.logger().warning(
                f"Changing position mode to {PositionMode.ONEWAY.name} did not succeed.")
            self._position_mode_ready = False

    def did_change_position_mode_fail(self, position_mode_changed_event: PositionModeChangeEvent):
        self.logger().error(
            f"Changing position mode to {PositionMode.ONEWAY.name} failed. "
            f"Reason: {position_mode_changed_event.message}.")
        self._position_mode_ready = False
        self.logger().warning("Cannot continue. Please resolve the issue in the account.")

    def update_complete_order_id_lists(self, order_id: str):
        if self._strategy_state == StrategyState.Opening:
            self._completed_opening_order_ids.append(order_id)
        elif self._strategy_state == StrategyState.Closing:
            self._completed_closing_order_ids.append(order_id)



#!/usr/bin/env python
import json
from typing import TYPE_CHECKING, List, Optional, TypedDict

from hummingbot.client.command.gateway_api_manager import begin_placeholder_mode
from hummingbot.core.gateway.gateway_http_client import GatewayStatus
from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


class PoolListInfo(TypedDict):
    """Pool information structure returned by gateway get_pool endpoint."""
    type: str  # "amm" or "clmm"
    network: str
    baseSymbol: str
    quoteSymbol: str
    address: str


def ensure_gateway_online(func):
    def wrapper(self, *args, **kwargs):
        if self.trading_core.gateway_monitor.gateway_status is GatewayStatus.OFFLINE:
            self.logger().error("Gateway is offline")
            return
        return func(self, *args, **kwargs)
    return wrapper


class GatewayPoolCommand:
    """Commands for managing gateway pools."""

    @ensure_gateway_online
    def gateway_pool(self, connector: Optional[str], trading_pair: Optional[str], action: Optional[str], args: List[str] = None):
        """
        View or update pool information.
        Usage:
            gateway pool <connector> <trading_pair>                - View pool information
            gateway pool <connector> <trading_pair> update         - Add/update pool information (interactive)
            gateway pool <connector> <trading_pair> update <address> - Add/update pool information (direct)
        """
        if args is None:
            args = []

        if not connector or not trading_pair:
            # Show help when insufficient arguments provided
            self.notify("\nGateway Pool Commands:")
            self.notify("  gateway pool <connector> <trading_pair>                - View pool information")
            self.notify("  gateway pool <connector> <trading_pair> update         - Add/update pool information (interactive)")
            self.notify("  gateway pool <connector> <trading_pair> update <address> - Add/update pool information (direct)")
            self.notify("\nExamples:")
            self.notify("  gateway pool uniswap/amm ETH-USDC")
            self.notify("  gateway pool raydium/clmm SOL-USDC update")
            self.notify("  gateway pool uniswap/amm ETH-USDC update 0x88e6a0c2ddd26feeb64f039a2c41296fcb3f5640")
            return

        if action == "update":
            if args and len(args) > 0:
                # Non-interactive mode: gateway pool <connector> <trading_pair> update <address>
                pool_address = args[0]
                safe_ensure_future(
                    self._update_pool_direct(connector, trading_pair, pool_address),
                    loop=self.ev_loop
                )
            else:
                # Interactive mode: gateway pool <connector> <trading_pair> update
                safe_ensure_future(
                    self._update_pool_interactive(connector, trading_pair),
                    loop=self.ev_loop
                )
        else:
            safe_ensure_future(
                self._view_pool(connector, trading_pair),
                loop=self.ev_loop
            )

    async def _view_pool(
        self,  # type: HummingbotApplication
        connector: str,
        trading_pair: str
    ):
        """View pool information."""
        try:
            # Parse connector format
            if "/" not in connector:
                self.notify(f"Error: Invalid connector format '{connector}'. Use format like 'uniswap/amm'")
                return

            connector_parts = connector.split("/")
            connector_name = connector_parts[0]
            trading_type = connector_parts[1]

            # Parse trading pair
            if "-" not in trading_pair:
                self.notify(f"Error: Invalid trading pair format '{trading_pair}'. Use format like 'ETH-USDC'")
                return

            # Capitalize the trading pair
            trading_pair = trading_pair.upper()

            # Get chain and network from connector
            chain, network, error = await self._get_gateway_instance().get_connector_chain_network(
                connector
            )

            if error:
                self.notify(error)
                return

            self.notify(f"\nFetching pool information for {trading_pair} on {connector}...")

            # Get pool information
            response = await self._get_gateway_instance().get_pool(
                trading_pair=trading_pair,
                connector=connector_name,
                network=network,
                type=trading_type
            )

            if "error" in response:
                self.notify(f"\nError: {response['error']}")
                self.notify(f"Pool {trading_pair} not found on {connector}")
                self.notify(f"You may need to add it using 'gateway pool {connector} {trading_pair} update'")
            else:
                # Display pool information
                try:
                    GatewayPoolCommand._display_pool_info(self, response, connector, trading_pair)
                except Exception as display_error:
                    # Log the response structure for debugging
                    self.notify(f"\nReceived pool data: {response}")
                    self.notify(f"Error displaying pool information: {str(display_error)}")

        except Exception as e:
            self.notify(f"Error fetching pool information: {str(e)}")

    async def _update_pool_direct(
        self,  # type: HummingbotApplication
        connector: str,
        trading_pair: str,
        pool_address: str
    ):
        """Direct mode to add a pool with just the address."""
        try:
            # Parse connector format
            if "/" not in connector:
                self.notify(f"Error: Invalid connector format '{connector}'. Use format like 'uniswap/amm'")
                return

            connector_parts = connector.split("/")
            connector_name = connector_parts[0]
            trading_type = connector_parts[1]

            # Parse trading pair
            if "-" not in trading_pair:
                self.notify(f"Error: Invalid trading pair format '{trading_pair}'. Use format like 'ETH-USDC'")
                return

            # Capitalize the trading pair
            trading_pair = trading_pair.upper()

            # Get chain and network from connector
            chain, network, error = await self._get_gateway_instance().get_connector_chain_network(
                connector
            )

            if error:
                self.notify(error)
                return

            self.notify(f"\nAdding pool for {trading_pair} on {connector}")
            self.notify(f"Chain: {chain}")
            self.notify(f"Network: {network}")

            # Fetch pool info from Gateway
            self.notify("\nFetching pool information from Gateway...")
            try:
                pool_info_response = await self._get_gateway_instance().pool_info(
                    connector=connector,
                    network=network,
                    pool_address=pool_address
                )

                if "error" in pool_info_response:
                    self.notify(f"Error fetching pool info: {pool_info_response['error']}")
                    self.notify("Cannot add pool without valid pool information")
                    return

                # Extract pool information from response
                fetched_base_symbol = pool_info_response.get("baseSymbol")
                fetched_quote_symbol = pool_info_response.get("quoteSymbol")
                base_token_address = pool_info_response.get("baseTokenAddress")
                quote_token_address = pool_info_response.get("quoteTokenAddress")
                fee_pct = pool_info_response.get("feePct")

                # If symbols are missing, try to fetch them from token addresses
                if not fetched_base_symbol and base_token_address:
                    try:
                        base_token_info = await self._get_gateway_instance().get_token(
                            symbol_or_address=base_token_address,
                            chain=chain,
                            network=network
                        )
                        if "token" in base_token_info and "symbol" in base_token_info["token"]:
                            fetched_base_symbol = base_token_info["token"]["symbol"]
                    except Exception:
                        # Silently skip - symbols are optional
                        pass

                if not fetched_quote_symbol and quote_token_address:
                    try:
                        quote_token_info = await self._get_gateway_instance().get_token(
                            symbol_or_address=quote_token_address,
                            chain=chain,
                            network=network
                        )
                        if "token" in quote_token_info and "symbol" in quote_token_info["token"]:
                            fetched_quote_symbol = quote_token_info["token"]["symbol"]
                    except Exception:
                        # Silently skip - symbols are optional
                        pass

                # Show warning if symbols couldn't be fetched
                if not fetched_base_symbol or not fetched_quote_symbol:
                    self.notify("\n⚠️  Warning: Could not determine token symbols from pool")
                    if not fetched_base_symbol:
                        self.notify(f"  - Base token symbol unknown (address: {base_token_address})")
                    if not fetched_quote_symbol:
                        self.notify(f"  - Quote token symbol unknown (address: {quote_token_address})")
                    self.notify("  Pool will be added without symbols")

                # Display fetched pool information
                self.notify("\n=== Pool Information ===")
                self.notify(f"Connector: {connector}")
                if fetched_base_symbol and fetched_quote_symbol:
                    self.notify(f"Trading Pair: {fetched_base_symbol}-{fetched_quote_symbol}")
                self.notify(f"Pool Type: {trading_type}")
                self.notify(f"Network: {network}")
                self.notify(f"Base Token: {fetched_base_symbol if fetched_base_symbol else 'N/A'}")
                self.notify(f"Quote Token: {fetched_quote_symbol if fetched_quote_symbol else 'N/A'}")
                self.notify(f"Base Token Address: {base_token_address}")
                self.notify(f"Quote Token Address: {quote_token_address}")
                if fee_pct is not None:
                    self.notify(f"Fee: {fee_pct}%")
                self.notify(f"Pool Address: {pool_address}")

                # Create pool data with required and optional fields
                pool_data = {
                    "address": pool_address,
                    "type": trading_type,
                    "baseTokenAddress": base_token_address,
                    "quoteTokenAddress": quote_token_address
                }
                # Add optional fields
                if fetched_base_symbol:
                    pool_data["baseSymbol"] = fetched_base_symbol
                if fetched_quote_symbol:
                    pool_data["quoteSymbol"] = fetched_quote_symbol
                if fee_pct is not None:
                    pool_data["feePct"] = fee_pct

                # Display pool data that will be stored
                self.notify("\nPool to add:")
                self.notify(json.dumps(pool_data, indent=2))

            except Exception as e:
                self.notify(f"Error fetching pool information: {str(e)}")
                self.notify("Cannot add pool without valid pool information")
                return

            # Add pool
            self.notify("\nAdding pool...")
            result = await self._get_gateway_instance().add_pool(
                connector=connector_name,
                network=network,
                pool_data=pool_data
            )

            if "error" in result:
                self.notify(f"Error: {result['error']}")
            else:
                self.notify("✓ Pool successfully added!")

                # Restart gateway for changes to take effect
                self.notify("\nRestarting Gateway for changes to take effect...")
                try:
                    await self._get_gateway_instance().post_restart()
                    self.notify("✓ Gateway restarted successfully")
                    self.notify(f"\nPool has been added. You can view it with: gateway pool {connector} {trading_pair}")
                except Exception as e:
                    self.notify(f"⚠️  Failed to restart Gateway: {str(e)}")
                    self.notify("You may need to restart Gateway manually for changes to take effect")

        except Exception as e:
            self.notify(f"Error adding pool: {str(e)}")

    async def _update_pool_interactive(
        self,  # type: HummingbotApplication
        connector: str,
        trading_pair: str
    ):
        """Interactive flow to add a pool."""
        try:
            # Parse connector format
            if "/" not in connector:
                self.notify(f"Error: Invalid connector format '{connector}'. Use format like 'uniswap/amm'")
                return

            connector_parts = connector.split("/")
            connector_name = connector_parts[0]
            trading_type = connector_parts[1]

            # Parse trading pair
            if "-" not in trading_pair:
                self.notify(f"Error: Invalid trading pair format '{trading_pair}'. Use format like 'ETH-USDC'")
                return

            # Capitalize the trading pair
            trading_pair = trading_pair.upper()

            # Get chain and network from connector
            chain, network, error = await self._get_gateway_instance().get_connector_chain_network(
                connector
            )

            if error:
                self.notify(error)
                return

            self.notify(f"\n=== Add Pool for {trading_pair} on {connector} ===")
            self.notify(f"Chain: {chain}")
            self.notify(f"Network: {network}")

            with begin_placeholder_mode(self):
                # Check if pool already exists
                try:
                    existing_pool = await self._get_gateway_instance().get_pool(
                        trading_pair=trading_pair,
                        connector=connector_name,
                        network=network,
                        type=trading_type
                    )
                except Exception:
                    # Pool doesn't exist, which is fine for adding a new pool
                    existing_pool = {"error": "Pool not found"}

                if "error" not in existing_pool:
                    # Pool exists, show current info
                    self.notify("\nPool already exists:")
                    GatewayPoolCommand._display_pool_info(self, existing_pool, connector, trading_pair)

                    # Ask if they want to update
                    response = await self.app.prompt(
                        prompt="Do you want to update this pool? (Yes/No) >>> "
                    )

                    if response.lower() not in ["y", "yes"]:
                        self.notify("Pool update cancelled")
                        return
                else:
                    self.notify(f"\nPool '{trading_pair}' not found. Let's add it to {chain} ({network}).")

                # Collect pool information
                self.notify("\nEnter pool information:")

                # Pool address
                pool_address = await self.app.prompt(
                    prompt="Pool contract address: "
                )
                if self.app.to_stop_config or not pool_address:
                    self.notify("Pool addition cancelled")
                    return

                # Fetch pool info from Gateway
                self.notify("\nFetching pool information from Gateway...")
                try:
                    pool_info_response = await self._get_gateway_instance().pool_info(
                        connector=connector,
                        network=network,
                        pool_address=pool_address
                    )

                    if "error" in pool_info_response:
                        self.notify(f"Error fetching pool info: {pool_info_response['error']}")
                        self.notify("Cannot add pool without valid pool information")
                        return

                    # Extract pool information from response
                    fetched_base_symbol = pool_info_response.get("baseSymbol")
                    fetched_quote_symbol = pool_info_response.get("quoteSymbol")
                    base_token_address = pool_info_response.get("baseTokenAddress")
                    quote_token_address = pool_info_response.get("quoteTokenAddress")
                    fee_pct = pool_info_response.get("feePct")

                    # If symbols are missing, try to fetch them from token addresses
                    if not fetched_base_symbol and base_token_address:
                        try:
                            base_token_info = await self._get_gateway_instance().get_token(
                                symbol_or_address=base_token_address,
                                chain=chain,
                                network=network
                            )
                            if "token" in base_token_info and "symbol" in base_token_info["token"]:
                                fetched_base_symbol = base_token_info["token"]["symbol"]
                        except Exception:
                            # Silently skip - symbols are optional
                            pass

                    if not fetched_quote_symbol and quote_token_address:
                        try:
                            quote_token_info = await self._get_gateway_instance().get_token(
                                symbol_or_address=quote_token_address,
                                chain=chain,
                                network=network
                            )
                            if "token" in quote_token_info and "symbol" in quote_token_info["token"]:
                                fetched_quote_symbol = quote_token_info["token"]["symbol"]
                        except Exception:
                            # Silently skip - symbols are optional
                            pass

                    # Show warning if symbols couldn't be fetched
                    if not fetched_base_symbol or not fetched_quote_symbol:
                        self.notify("\n⚠️  Warning: Could not determine token symbols from pool")
                        if not fetched_base_symbol:
                            self.notify(f"  - Base token symbol unknown (address: {base_token_address})")
                        if not fetched_quote_symbol:
                            self.notify(f"  - Quote token symbol unknown (address: {quote_token_address})")
                        self.notify("  Pool will be added without symbols")

                    # Display fetched pool information
                    self.notify("\n=== Pool Information ===")
                    self.notify(f"Connector: {connector}")
                    if fetched_base_symbol and fetched_quote_symbol:
                        self.notify(f"Trading Pair: {fetched_base_symbol}-{fetched_quote_symbol}")
                    self.notify(f"Pool Type: {trading_type}")
                    self.notify(f"Network: {network}")
                    self.notify(f"Base Token: {fetched_base_symbol if fetched_base_symbol else 'N/A'}")
                    self.notify(f"Quote Token: {fetched_quote_symbol if fetched_quote_symbol else 'N/A'}")
                    self.notify(f"Base Token Address: {base_token_address}")
                    self.notify(f"Quote Token Address: {quote_token_address}")
                    if fee_pct is not None:
                        self.notify(f"Fee: {fee_pct}%")
                    self.notify(f"Pool Address: {pool_address}")

                    # Create pool data with required and optional fields
                    pool_data = {
                        "address": pool_address,
                        "type": trading_type,
                        "baseTokenAddress": base_token_address,
                        "quoteTokenAddress": quote_token_address
                    }
                    # Add optional fields
                    if fetched_base_symbol:
                        pool_data["baseSymbol"] = fetched_base_symbol
                    if fetched_quote_symbol:
                        pool_data["quoteSymbol"] = fetched_quote_symbol
                    if fee_pct is not None:
                        pool_data["feePct"] = fee_pct

                    # Display pool data that will be stored
                    self.notify("\nPool to add:")
                    self.notify(json.dumps(pool_data, indent=2))

                    # Confirm
                    confirm = await self.app.prompt(
                        prompt="Add this pool? (Yes/No) >>> "
                    )

                    if confirm.lower() not in ["y", "yes"]:
                        self.notify("Pool addition cancelled")
                        return

                except Exception as e:
                    self.notify(f"Error fetching pool information: {str(e)}")
                    self.notify("Cannot add pool without valid pool information")
                    return

                # Add pool
                self.notify("\nAdding pool...")
                result = await self._get_gateway_instance().add_pool(
                    connector=connector_name,
                    network=network,
                    pool_data=pool_data
                )

                if "error" in result:
                    self.notify(f"Error: {result['error']}")
                else:
                    self.notify("✓ Pool successfully added!")

                    # Restart gateway for changes to take effect
                    self.notify("\nRestarting Gateway for changes to take effect...")
                    try:
                        await self._get_gateway_instance().post_restart()
                        self.notify("✓ Gateway restarted successfully")
                        self.notify(f"\nPool has been added. You can view it with: gateway pool {connector} {trading_pair}")
                    except Exception as e:
                        self.notify(f"⚠️  Failed to restart Gateway: {str(e)}")
                        self.notify("You may need to restart Gateway manually for changes to take effect")

        except Exception as e:
            self.notify(f"Error updating pool: {str(e)}")

    def _display_pool_info(
        self,
        pool_info: dict,
        connector: str,
        trading_pair: str
    ):
        """Display pool information in a formatted way."""
        self.notify("\n=== Pool Information ===")
        self.notify(f"Connector: {connector}")
        self.notify(f"Trading Pair: {trading_pair}")
        self.notify(f"Pool Type: {pool_info.get('type', 'N/A')}")
        self.notify(f"Network: {pool_info.get('network', 'N/A')}")
        self.notify(f"Base Token: {pool_info.get('baseSymbol', 'N/A')}")
        self.notify(f"Quote Token: {pool_info.get('quoteSymbol', 'N/A')}")
        self.notify(f"Base Token Address: {pool_info.get('baseTokenAddress', 'N/A')}")
        self.notify(f"Quote Token Address: {pool_info.get('quoteTokenAddress', 'N/A')}")
        self.notify(f"Fee: {pool_info.get('feePct', 'N/A')}%")
        self.notify(f"Pool Address: {pool_info.get('address', 'N/A')}")



import numbers
from decimal import Decimal
from random import randint
from typing import Any, Dict, Optional

from pydantic import ConfigDict, Field, SecretStr

import hummingbot.connector.exchange.vertex.vertex_constants as CONSTANTS
from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

CENTRALIZED = True
USE_ETHEREUM_WALLET = False
EXAMPLE_PAIR = "WBTC-USDC"
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.0"),
    taker_percent_fee_decimal=Decimal("0.0002"),
)


def hex_to_bytes32(hex_string: str) -> bytes:
    if hex_string.startswith("0x"):
        hex_string = hex_string[2:]
    data_bytes = bytes.fromhex(hex_string)
    padded_data = data_bytes + b"\x00" * (32 - len(data_bytes))
    return padded_data


def convert_timestamp(timestamp: Any) -> float:
    return float(timestamp) / 1e9


def trading_pair_to_product_id(trading_pair: str, exchange_market_info: Dict, is_perp: Optional[bool] = False) -> int:
    tp = trading_pair.replace("-", "/")
    for product_id in exchange_market_info:
        if is_perp and "perp" not in exchange_market_info[product_id]["symbol"].lower():
            continue
        if exchange_market_info[product_id]["market"] == tp:
            return product_id
    return -1


def market_to_trading_pair(market: str) -> str:
    """Converts a market symbol from Vertex to a trading pair."""
    return market.replace("/", "-")


def convert_from_x18(data: Any, precision: Optional[Decimal] = None) -> Any:
    """
    Converts numerical data encoded as x18 to a string representation of a
    floating point number, recursively applies the conversion for other data types.
    """
    if data is None:
        return None

    # Check if data type is str or float
    if isinstance(data, str) or isinstance(data, numbers.Number):
        data = Decimal(data) / Decimal("1000000000000000000")  # type: ignore
        if precision:
            data = data.quantize(precision)
        return str(data)

    if isinstance(data, dict):
        for k, v in data.items():
            data[k] = convert_from_x18(v, precision)
    elif isinstance(data, list):
        for i in range(0, len(data)):
            data[i] = convert_from_x18(data[i], precision)
    else:
        raise TypeError("Data is of unsupported type for convert_from_x18 to process", data)
    return data


def convert_to_x18(data: Any, precision: Optional[Decimal] = None) -> Any:
    """
    Converts numerical data encoded to a string representation of x18, recursively
    applies the conversion for other data types.
    """
    if data is None:
        return None

    # Check if data type is str or float
    if isinstance(data, str) or isinstance(data, numbers.Number):
        data = Decimal(str(data))  # type: ignore
        if precision:
            data = data.quantize(precision)
        return str((data * Decimal("1000000000000000000")).quantize(Decimal("1")))

    if isinstance(data, dict):
        for k, v in data.items():
            data[k] = convert_to_x18(v, precision)
    elif isinstance(data, list):
        for i in range(0, len(data)):
            data[i] = convert_to_x18(data[i], precision)
    else:
        raise TypeError("Data is of unsupported type for convert_to_x18 to process", data)
    return data


def generate_expiration(timestamp: float = None, order_type: Optional[str] = None) -> str:
    default_max_time = 8640000000000000  # NOTE: Forever
    default_day_time = 86400
    # Default significant bit is 0 for GTC
    # https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/place-order
    sig_bit = 0

    if order_type == CONSTANTS.TIME_IN_FORCE_IOC:
        sig_bit = 1
    elif order_type == CONSTANTS.TIME_IN_FORCE_FOK:
        sig_bit = 2
    elif order_type == CONSTANTS.TIME_IN_FORCE_POSTONLY:
        sig_bit = 3

    # NOTE: We can setup maxtime
    expiration = str(default_max_time | (sig_bit << 62))

    if timestamp:
        unix_epoch = int(timestamp)
        expiration = str((unix_epoch + default_day_time) | (sig_bit << 62))

    return expiration


def generate_nonce(timestamp: float, expiry_ms: int = 90) -> int:
    unix_epoch_ms = int((timestamp * 1000) + (expiry_ms * 1000))
    nonce = (unix_epoch_ms << 20) + randint(1, 1001)
    return nonce


def convert_address_to_sender(address: str) -> str:
    # NOTE: the sender address includes the subaccount, which is "default" by default, you cannot interact with
    # subaccounts outside of default on the UI currently.
    # https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes#signing
    if isinstance(address, str):
        default_12bytes = "64656661756c740000000000"
        return address + default_12bytes
    raise TypeError("Address must be of type string")


def is_exchange_information_valid(exchange_info: Dict[str, Any]) -> bool:
    """
    Default's to true, there isn't anything to check agaisnt.
    """
    return True


class VertexConfigMap(BaseConnectorConfigMap):
    connector: str = "vertex"
    vertex_arbitrum_private_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Arbitrum private key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    vertex_arbitrum_address: str = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Arbitrum wallet address",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="vertex")


KEYS = VertexConfigMap.model_construct()


class VertexTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = "vertex_testnet"
    vertex_testnet_arbitrum_private_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Arbitrum TESTNET private key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    vertex_testnet_arbitrum_address: str = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Arbitrum TESTNET wallet address",
            "is_secure": False,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="vertex_testnet")


OTHER_DOMAINS = ["vertex_testnet"]
OTHER_DOMAINS_PARAMETER = {"vertex_testnet": "vertex_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"vertex_testnet": "WBTC-USDC"}
OTHER_DOMAINS_DEFAULT_FEES = {"vertex_testnet": DEFAULT_FEES}

OTHER_DOMAINS_KEYS = {"vertex_testnet": VertexTestnetConfigMap.model_construct()}



import asyncio
import logging
from datetime import datetime, timezone
from typing import List, Optional

from dateutil.parser import parse as dateparse

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.data_feed.candles_feed.btc_markets_spot_candles import constants as CONSTANTS
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.logger import HummingbotLogger


class BtcMarketsSpotCandles(CandlesBase):
    """
    BTC Markets implementation for fetching candlestick data.

    Note: BTC Markets doesn't support WebSocket for candles, so we use constant polling.
    This implementation maintains a constant polling rate to capture real-time updates
    and fills gaps with heartbeat candles to maintain equidistant intervals.
    """

    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)

        self._consecutive_empty_responses = 0
        self._historical_fill_in_progress = False

        # Task management for polling
        self._polling_task: Optional[asyncio.Task] = None
        self._shutdown_event = asyncio.Event()
        self._is_running = False

    @property
    def name(self):
        return f"btc_markets_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        # BTC Markets doesn't support WebSocket for candles
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        market_id = self.get_exchange_trading_pair(self._trading_pair)
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT.format(market_id=market_id)

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    @property
    def _last_real_candle(self):
        """Get the last candle, filtering out heartbeats if needed."""
        if not self._candles:
            return None
        # Find last candle with volume > 0, or just return last candle
        for candle in reversed(self._candles):
            if candle[5] > 0:  # volume > 0
                return candle
        return self._candles[-1]

    @property
    def _current_candle_timestamp(self):
        return self._candles[-1][0] if self._candles else None

    async def start_network(self):
        """
        Start the network and begin polling.
        """
        await self.stop_network()
        await self.initialize_exchange_data()
        self._is_running = True
        self._shutdown_event.clear()
        self._polling_task = asyncio.create_task(self._polling_loop())

    async def stop_network(self):
        """
        Stop the network by gracefully shutting down the polling task.
        """
        if self._polling_task and not self._polling_task.done():
            self._is_running = False
            self._shutdown_event.set()

            try:
                # Wait for graceful shutdown
                await asyncio.wait_for(self._polling_task, timeout=10.0)
            except asyncio.TimeoutError:
                self.logger().warning("Polling task didn't stop gracefully, cancelling...")
                self._polling_task.cancel()
                try:
                    await self._polling_task
                except asyncio.CancelledError:
                    pass

        self._polling_task = None
        self._is_running = False

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(
            url=self.health_check_url, throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT
        )
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        """
        Converts from the Hummingbot trading pair format to the exchange's trading pair format.
        BTC Markets uses the same format so no conversion is needed.
        """
        return trading_pair

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(
        self, start_time: Optional[int] = None, end_time: Optional[int] = None, limit: Optional[int] = None
    ) -> dict:
        """
        Generates parameters for the REST API request to fetch candles.
        """
        params = {
            "timeWindow": self.intervals[self.interval],
        }

        if start_time is None and end_time is None:
            # For real-time polling, fetch a small number of recent candles
            params["limit"] = limit if limit is not None else 3
        else:
            # Use timestamp parameters for historical data
            params["limit"] = min(limit if limit is not None else 1000, 1000)

            if start_time:
                start_iso = datetime.fromtimestamp(start_time, tz=timezone.utc).isoformat().replace("+00:00", "Z")
                params["from"] = start_iso

            if end_time:
                end_iso = datetime.fromtimestamp(end_time, tz=timezone.utc).isoformat().replace("+00:00", "Z")
                params["to"] = end_iso

        return params

    def _parse_rest_candles(self, data: List[List[str]], end_time: Optional[int] = None) -> List[List[float]]:
        """
        Parse the REST API response into the standard candle format.
        """
        if not isinstance(data, list) or len(data) == 0:
            return []

        new_hb_candles = []
        for i, candle in enumerate(data):
            try:
                if not isinstance(candle, list) or len(candle) < 6:
                    self.logger().warning(f"Invalid candle format at index {i}: {candle}")
                    continue

                timestamp = self.ensure_timestamp_in_seconds(dateparse(candle[0]).timestamp())
                open_price = float(candle[1])
                high = float(candle[2])
                low = float(candle[3])
                close = float(candle[4])
                volume = float(candle[5])

                # BTC Markets doesn't provide these values
                quote_asset_volume = 0.0
                n_trades = 0.0
                taker_buy_base_volume = 0.0
                taker_buy_quote_volume = 0.0

                new_hb_candles.append(
                    [
                        timestamp,
                        open_price,
                        high,
                        low,
                        close,
                        volume,
                        quote_asset_volume,
                        n_trades,
                        taker_buy_base_volume,
                        taker_buy_quote_volume,
                    ]
                )

            except Exception as e:
                self.logger().error(f"Error parsing candle {candle}: {e}")

        # Sort by timestamp (oldest first)
        new_hb_candles.sort(key=lambda x: x[0])
        return new_hb_candles

    def _create_heartbeat_candle(self, timestamp: float) -> List[float]:
        """
        Create a "heartbeat" candle for periods with no trading activity.
        Uses the close price from the last real candle.
        """
        last_real = self._last_real_candle
        if last_real is not None:
            close_price = last_real[4]
        elif self._candles:
            close_price = self._candles[-1][4]
        else:
            close_price = 0.0

        return [timestamp, close_price, close_price, close_price, close_price, 0.0, 0.0, 0.0, 0.0, 0.0]

    def _fill_gaps_and_append(self, new_candle: List[float]):
        """
        Fill any gaps between last candle and new candle, then append the new candle.
        """
        if not self._candles:
            self._candles.append(new_candle)
            return

        last_timestamp = self._candles[-1][0]
        new_timestamp = new_candle[0]

        # Fill gaps with heartbeats
        current_timestamp = last_timestamp + self.interval_in_seconds
        while current_timestamp < new_timestamp:
            heartbeat = self._create_heartbeat_candle(current_timestamp)
            self._candles.append(heartbeat)
            self.logger().debug(f"Added heartbeat candle at {current_timestamp}")
            current_timestamp += self.interval_in_seconds

        # Append the new candle
        self._candles.append(new_candle)
        self.logger().debug(f"Added new candle at {new_timestamp}")

    def _ensure_heartbeats_to_current_time(self):
        """
        Ensure we have heartbeat candles up to the current time interval.
        Only creates heartbeats for complete intervals (not the current incomplete one).
        """
        if not self._candles:
            return

        current_time = self._time()
        current_interval_timestamp = self._round_timestamp_to_interval_multiple(current_time)
        last_candle_timestamp = self._candles[-1][0]

        # Only create heartbeats for complete intervals
        next_expected_timestamp = last_candle_timestamp + self.interval_in_seconds

        while next_expected_timestamp < current_interval_timestamp:
            heartbeat = self._create_heartbeat_candle(next_expected_timestamp)
            self._candles.append(heartbeat)
            self.logger().debug(f"Added heartbeat for time progression: {next_expected_timestamp}")
            next_expected_timestamp += self.interval_in_seconds

    async def fill_historical_candles(self):
        """
        Fill historical candles with heartbeats to maintain equidistant intervals.
        """
        if self._historical_fill_in_progress:
            return

        self._historical_fill_in_progress = True

        try:
            iteration = 0
            max_iterations = 20

            while not self.ready and len(self._candles) > 0 and iteration < max_iterations:
                iteration += 1

                try:
                    oldest_timestamp = self._candles[0][0]
                    missing_records = self._candles.maxlen - len(self._candles)

                    if missing_records <= 0:
                        break

                    end_timestamp = oldest_timestamp - self.interval_in_seconds
                    start_timestamp = end_timestamp - (missing_records * self.interval_in_seconds)

                    # Fetch real candles for this time range
                    real_candles = await self.fetch_candles(
                        start_time=start_timestamp, end_time=end_timestamp + self.interval_in_seconds
                    )

                    # Fill gaps with heartbeats
                    complete_candles = self._fill_historical_gaps_with_heartbeats(
                        real_candles, start_timestamp, end_timestamp
                    )

                    if complete_candles:
                        candles_to_add = (
                            complete_candles[-missing_records:]
                            if len(complete_candles) > missing_records
                            else complete_candles
                        )

                        # Add in reverse order to maintain chronological order
                        for candle in reversed(candles_to_add):
                            self._candles.appendleft(candle)
                    else:
                        break

                    await self._sleep(0.1)

                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    self.logger().exception(f"Error during historical fill iteration {iteration}: {e}")
                    await self._sleep(1.0)

        finally:
            self._historical_fill_in_progress = False

    def _fill_historical_gaps_with_heartbeats(
        self, candles: List[List[float]], start_timestamp: float, end_timestamp: float
    ) -> List[List[float]]:
        """
        Fill gaps in historical candle data with heartbeat candles.
        """
        if not candles.any():
            # Generate all heartbeats
            result = []
            current_timestamp = self._round_timestamp_to_interval_multiple(start_timestamp)
            interval_count = 0

            while current_timestamp <= end_timestamp and interval_count < 1000:
                heartbeat = self._create_heartbeat_candle(current_timestamp)
                result.append(heartbeat)
                current_timestamp += self.interval_in_seconds
                interval_count += 1

            return result

        # Create map of real candles by timestamp
        candle_map = {self._round_timestamp_to_interval_multiple(c[0]): c for c in candles}

        # Fill complete time range
        result = []
        current_timestamp = self._round_timestamp_to_interval_multiple(start_timestamp)
        interval_count = 0

        while current_timestamp <= end_timestamp and interval_count < 1000:
            if current_timestamp in candle_map:
                result.append(candle_map[current_timestamp])
            else:
                heartbeat = self._create_heartbeat_candle(current_timestamp)
                result.append(heartbeat)

            current_timestamp += self.interval_in_seconds
            interval_count += 1

        return result

    async def fetch_recent_candles(self, limit: int = 3) -> List[List[float]]:
        """Fetch recent candles from the API."""
        try:
            params = {"timeWindow": self.intervals[self.interval], "limit": limit}

            rest_assistant = await self._api_factory.get_rest_assistant()
            response = await rest_assistant.execute_request(
                url=self.candles_url,
                throttler_limit_id=self._rest_throttler_limit_id,
                params=params,
                method=self._rest_method,
            )

            return self._parse_rest_candles(response)

        except Exception as e:
            self.logger().error(f"Error fetching recent candles: {e}")
            return []

    async def _polling_loop(self):
        """
        Main polling loop - separated from listen_for_subscriptions for better testability.
        This method can be cancelled cleanly and tested independently.
        """
        try:
            self.logger().info(f"Starting constant polling for {self._trading_pair} candles")

            # Initial setup
            await self._initialize_candles()

            while self._is_running and not self._shutdown_event.is_set():
                try:
                    # Poll for updates
                    await self._poll_and_update_candles()

                    # Ensure heartbeats up to current time
                    self._ensure_heartbeats_to_current_time()

                    # Wait for either shutdown signal or polling interval
                    try:
                        await asyncio.wait_for(
                            self._shutdown_event.wait(),
                            timeout=CONSTANTS.POLL_INTERVAL
                        )
                        # If we reach here, shutdown was requested
                        break
                    except asyncio.TimeoutError:
                        # Normal case - polling interval elapsed
                        continue

                except asyncio.CancelledError:
                    self.logger().info("Polling loop cancelled")
                    raise
                except Exception as e:
                    self.logger().exception(f"Unexpected error during polling: {e}")

                    # Wait before retrying, but also listen for shutdown
                    try:
                        await asyncio.wait_for(
                            self._shutdown_event.wait(),
                            timeout=5.0
                        )
                        break
                    except asyncio.TimeoutError:
                        continue

        finally:
            self.logger().info("Polling loop stopped")
            self._is_running = False

    async def listen_for_subscriptions(self):
        """
        Legacy method for compatibility with base class.
        Now just delegates to the task-based approach.
        """
        if not self._is_running:
            await self.start_network()

        # Wait for the polling task to complete
        if self._polling_task:
            try:
                await self._polling_task
            except asyncio.CancelledError:
                self.logger().info("Listen for subscriptions cancelled")
                raise

    async def _poll_and_update_candles(self):
        """
        Fetch recent candles and update data structure.
        This method is now easily testable in isolation.
        """
        try:
            # Always fetch recent candles to get current candle updates
            recent_candles = await self.fetch_recent_candles(limit=3)

            if not recent_candles:
                self._consecutive_empty_responses += 1
                return

            # Reset empty response counter
            self._consecutive_empty_responses = 0
            latest_candle = recent_candles[-1]

            if not self._candles:
                # First initialization
                self._candles.append(latest_candle)
                self._ws_candle_available.set()
                safe_ensure_future(self.fill_historical_candles())
                return

            # Simple logic: append if newer, update if same timestamp
            last_timestamp = self._candles[-1][0]
            latest_timestamp = latest_candle[0]

            if latest_timestamp > last_timestamp:
                # New candle - fill gaps and append
                self._fill_gaps_and_append(latest_candle)
            elif latest_timestamp == last_timestamp:
                # Update current candle
                old_candle = self._candles[-1]
                self._candles[-1] = latest_candle

                # Log significant changes
                if abs(old_candle[4] - latest_candle[4]) > 0.0001 or abs(old_candle[5] - latest_candle[5]) > 0.0001:
                    self.logger().debug(
                        f"Updated current candle: close {old_candle[4]:.4f} -> {latest_candle[4]:.4f}, "
                        f"volume {old_candle[5]:.4f} -> {latest_candle[5]:.4f}"
                    )

        except Exception as e:
            self.logger().error(f"Error during polling: {e}")
            self._consecutive_empty_responses += 1

    async def _initialize_candles(self):
        """Initialize with recent candle data and start constant polling."""
        try:
            self.logger().info("Initializing candles with recent data...")

            candles = await self.fetch_recent_candles(limit=2)

            if candles:
                latest_candle = candles[-1]
                self._candles.append(latest_candle)
                self._ws_candle_available.set()
                safe_ensure_future(self.fill_historical_candles())
                self.logger().info(f"Initialized with candle at {latest_candle[0]}")
            else:
                self.logger().warning("No recent candles found during initialization")

        except Exception as e:
            self.logger().error(f"Failed to initialize candles: {e}")

    def ws_subscription_payload(self):
        """Not used for BTC Markets since WebSocket is not supported for candles."""
        raise NotImplementedError("WebSocket not supported for BTC Markets candles")

    def _parse_websocket_message(self, data):
        """Not used for BTC Markets since WebSocket is not supported for candles."""
        raise NotImplementedError("WebSocket not supported for BTC Markets candles")
